from pydantic import BaseModel


class AbstractInOutType(BaseModel):
    """
    Abstract In/Out type
    """

    content: str


class MarkdownString(AbstractInOutType):
    """A parsed Markdown Formatted String.
    Feats :
    - ATX header formatting.
    - Remove base64 images
    """


class HTMLString(AbstractInOutType):
    """A HTML Formatted String"""



from collections import Counter
from itertools import groupby
import re

from thefuzz import fuzz  # type: ignore : no stubs

from .utils import PdfParserState
from .components import TocTitle


class PdfTocExtraction(PdfParserState):
    """Class intended to be used to extract the table of content
    of a document.
    Intended to be a component inherited by pdfParser => PdfParser(PdfTocExtraction)
    """

    def get_toc(self) -> list[TocTitle]:
        """Gets the table of content of a document.
        The process:
        - tries to find the toc in metadatas
        - if not found, tries to find a table of content in the document
        --- if found, tries to infer the level of header...
             --- ...with the x offset the headers in the toc, assuming lower level header are more to the right
             --- ...or trying to parse a schema, such as 1. 1.1, 1.1.a)
        - if not found, tries to infer header level with font size, assuming lower level have lower font size

        Returns:
            list[TocTitle]: the list of titles found
        """
        toc_from_metadata = self.get_toc_from_metadata()
        # Run anyway as table of content might also be in document
        # and we want to flag the lines that belong to it
        toc_from_document = self.get_toc_from_document()

        toc = toc_from_metadata if toc_from_metadata else toc_from_document
        if toc:
            self._set_block_issectiontitle_with_toc(toc)
        else:
            self._set_block_issectiontitle_with_fontsize()

        return toc

    def get_toc_from_metadata(self) -> list[TocTitle] | None:
        """Uses pymupdf.get_toc() to try to get the table of
        content of the document from the metadatas.
        If it exists, builds the table of content from it

        Returns:
            list[TocTitle] | None: The toc titles found in metadatas
        """
        toc = self.document.get_toc() if self.document is not None else []  # type: ignore : missing typing in pymupdf | Document.get_toc() -> list[tuple[int, str, int, dict[str, Any]]]
        if not toc:
            return

        return [
            TocTitle(text=item[1], level=item[0], page=item[2], source="metadata")
            for item in toc
        ]

    def get_toc_from_document(self) -> list[TocTitle]:
        """Tries to find a table of content in the document
        and parses it's titles. May return an ampty list
        if no table of content found

        Returns:
            list[TocTitle]: the titles found in the toc
        """
        toc = self._find_toc_titles()
        if toc:
            toc = PdfTocExtraction._infer_level_with_schema(toc)
            if not all(t.level for t in toc):
                toc = PdfTocExtraction._infer_level_with_offset(toc)

        return toc

    def _find_toc_titles(self) -> list[TocTitle] | None:
        """Tries to find the table of content based on regex matches.
        Handles the case where the title of the toc is on multiple lines
        Returns:
            list[TocTitle]: The potential titles of the table of content
                and their caracteristics
        """
        toc_pattern = re.compile(
            r"(.+?)(?:\s+)?[\.\_\-\s….]{5,}(?:\s+)?(?:[pP]\.\s*)?(\d+)"
        )

        toc_titles: list[TocTitle] = []
        for i, line in enumerate(self.lines):
            match = re.match(toc_pattern, line.text)
            # regex may match ZIP codes or phone numbers => check page is less than 3 numbers
            if match and len(match[2]) < 4:
                line.is_toc_element = True
                # multiline toc title
                if i > 2 and (
                    not self.lines[i - 1].is_toc_element
                    and self.lines[i - 2].is_toc_element
                ):
                    toc_text = self.lines[i - 1].text + match[1]
                    x_offset = self.lines[i - 1].origin.x  # type: ignore : missing typing in pymupdf | Point.x : float
                    self.lines[i - 1].is_toc_element = True
                # single line toc title
                else:
                    toc_text = match[1]
                    x_offset = line.origin.x  # type: ignore : missing typing in pymupdf | Point.x : float
                toc_titles.append(
                    TocTitle(
                        text=toc_text,
                        page=int(match[2]),
                        x_offset=int(x_offset),
                        source="regex",
                        source_page=line.page,
                    )
                )

        return toc_titles

    @staticmethod
    def _infer_level_with_offset(toc_titles: list[TocTitle]) -> list[TocTitle]:
        """Given on a list of toc_tiles found in the document,
        infer their level (h1, h2, ...) based on the x_offset
        of their bbox. We assume that a bigger x_offset means lower level
        (= text is more on the right of the document). Like so:
        A. H1 title
          a. H2 title
            1. H3 title

        Args:
            toc_titles (list[TocTitle]): list of toc titles returned by
                _find_toc_titles()

        Returns:
            list[TocTitle]: the toc titles, with the "level" value set
        """
        new_toc_titles: list[TocTitle] = []
        toc_elem_per_page = groupby(toc_titles, key=lambda x: x.source_page)
        for _, toc_elem_on_page in toc_elem_per_page:
            toc_elem_on_page = list(toc_elem_on_page)
            offsets: list[float] = sorted(
                list(set(t.x_offset for t in toc_elem_on_page))
            )
            for title in toc_elem_on_page:
                title.level = offsets.index(title.x_offset) + 1
                new_toc_titles.append(title)

        return new_toc_titles

    @staticmethod
    def _infer_level_with_schema(toc_titles: list[TocTitle]) -> list[TocTitle]:
        """Tries to infer the titles levels based on the
        schema of the title sections. Like so:
        1. H1 title
        1.1 H2 title
        1.1.1 H3 title

        Args:
            toc_titles (list[TocTitle]): The list of toc title

        Returns:
            list[TocTitle]: the list of toc title with the level set (if schema found)
        """
        h1_pattern = re.compile(r"^(\d+)[\s\.\)]*")
        h2_pattern = re.compile(r"^(\d+)[\s\.\)]+(\d+)")
        h3_pattern = re.compile(r"^(\d+)[\s\.\)]+(\d+)[\s\.\)]+(\d+)")

        for t in toc_titles:
            for pattern, level in zip((h3_pattern, h2_pattern, h1_pattern), (3, 2, 1)):
                match = re.match(pattern, t.text)
                if match:
                    t.level = level
                    break

        return toc_titles

    def _set_block_issectiontitle_with_toc(self, toc: list[TocTitle]) -> None:
        """After finding the toc, this modifies
        blocks among self.blocks attributes of the PdfParser
        to set which ones are the section's headers.
        """
        for title in toc:
            pages_to_look_on = range(title.page - 1, title.page + 2)
            blocks_to_consider = (
                block for block in self.blocks if block.page in pages_to_look_on
            )
            best_ratio, best_block = 0, None
            for block in blocks_to_consider:
                ratio = fuzz.ratio(  # type: ignore : missing typing in fuzz | fuzz.ratio(s1 : str, s2: str) -> int
                    title.text.lower(), block.text.lower()
                )
                if ratio > best_ratio:
                    best_ratio = ratio
                    best_block = block
            if best_block and best_ratio >= 75:
                title.found = True
                best_block.section_title = title

    def _set_block_issectiontitle_with_fontsize(self) -> None:
        """Uses the fontize attribute of blocks to assign
        them a header level.
        We assume that bigger fontsize means higher level header.
        All block having a fontisze greater than the fontsize of body
        will be assigned a level based on their fontsize.
        """
        fontsize_counts = Counter(
            block.fontsize for block in self.blocks if not block.is_empty
        )
        if fontsize_counts:
            main_body_fontsize = max(fontsize_counts, key=fontsize_counts.get)
            fontsizes = (
                fontsize
                for fontsize in fontsize_counts.keys()
                if fontsize > main_body_fontsize
            )
            fontsizes = sorted(fontsizes, reverse=True)[:5]
            for block in self.blocks:
                if block.fontsize in fontsizes:
                    block.section_title = TocTitle(
                        text=block.text,
                        level=fontsizes.index(block.fontsize) + 1,
                        page=block.page,
                        source="fontsize",
                    )

    def _get_document_main_title(self) -> str:
        """
        Tries to infer the main title of the document.
        It considers the blocks belong to main title if
        they are on first page and their fontsize is bigger
        that the body size of the document.

        Returns :
            (str) : the main title of the document
        """
        fontsize_counts = Counter(
            span.fontsize for span in self.spans if not span.is_empty
        )
        if fontsize_counts:
            main_body_fontsize = max(fontsize_counts, key=fontsize_counts.get)
            spans_on_first_page = [
                s for s in self.spans if s.page == 0 and not s.is_empty
            ]

            if spans_on_first_page:
                # Get the 2 biggest fontsizes of 1st page
                first_page_biggest_fontsizes = sorted(
                    Counter(
                        span.fontsize
                        for span in spans_on_first_page
                        if not span.is_empty
                    ),
                    reverse=True,
                )[:2]
                title_spans = (
                    s
                    for s in spans_on_first_page
                    if s.fontsize > main_body_fontsize * 1.1
                    and s.fontsize in first_page_biggest_fontsizes
                )
                return " ".join((s.text for s in title_spans))[:100].strip()

        return ""



from itertools import groupby
from typing import Literal

import matplotlib.pyplot as plt
import numpy as np
import pymupdf  # type: ignore : not stubs

from .utils import PdfParserState
from .extract_tables import PdfTable
from .components import TextSpan, TextLine, TextBlock


class PdfPlotter(PdfParserState):
    """Tool that groups methods related to plotting elements obtained by
    the PdfParser. Meant to be as inerited class PdfParser(PdfPlotter)

    Most methods make use of PdfParser's attributes, such as:
    self.spans, self.lines, self.blocks, self.tables and self.document.
    """

    def plot_pdf(
        self,
        items_to_plot: list[Literal["span", "line", "block", "table"]] | None = None,
        dpi: int = 100,
    ) -> None:
        """Buils a plot of the pdf file.
        Plots the span's bboxes and tables grid

        Args:
            items_to_plot (list[str]) : the list of types of elements to plot.
                Can be any subset of ["span", "line", "block", "table"].
            dpi (int): the resolution of the generated image
        """
        if not items_to_plot:
            items_to_plot = ["span", "line", "block", "table"]

        ## copy pdf inmemory to draw on it
        pdfdata = self.document.tobytes()  # type: ignore : missing typing in pymupdf | Document.tobytes(*args) -> bytes
        doc_to_draw_on = pymupdf.open(stream=pdfdata, filetype="pdf")

        spans_per_page: dict[int, list[TextSpan]] = PdfPlotter._get_items_per_page(
            self.spans
        )
        lines_per_page: dict[int, list[TextLine]] = PdfPlotter._get_items_per_page(
            self.lines
        )
        blocks_per_page: dict[int, list[TextBlock]] = PdfPlotter._get_items_per_page(
            self.blocks
        )
        tables_per_page: dict[int, list[PdfTable]] = PdfPlotter._get_items_per_page(
            self.tables
        )

        for page_n in spans_per_page.keys():
            page = doc_to_draw_on.load_page(page_n)  # type: ignore : missing typing in pymupdf | Document.load_page(page_id: int) -> Page
            if "span" in items_to_plot and page_n in spans_per_page:
                for span in spans_per_page[page_n]:
                    style = PdfPlotter._get_rect_style(span)
                    page.draw_rect(span.bbox, **style, fill_opacity=0.5)  # type: ignore : missing typing in pymupdf
            if "line" in items_to_plot and page_n in lines_per_page:
                for line in lines_per_page[page_n]:
                    page.draw_rect(line.bbox, color=pymupdf.pdfcolor["black"], width=1)  # type: ignore : missing typing in pymupdf
            if "block" in items_to_plot and page_n in blocks_per_page:
                for block in blocks_per_page[page_n]:
                    page.draw_rect(  # type: ignore : missing typing in pymupdf
                        block.bbox,
                        color=pymupdf.pdfcolor["black"],  # type: ignore : missing typing in pymupdf | pymupdf.pdfcolor : dict[str, tuple[int, int, int]]
                        width=2,
                        dashes="[3 4] 0",
                    )
            if "table" in items_to_plot and page_n in tables_per_page:
                for table in tables_per_page[page_n]:
                    for cell in table.cells:
                        page.draw_rect(cell, color=pymupdf.pdfcolor["blue"])  # type: ignore : missing typing in pymupdf
                    page.draw_rect(table.bbox, color=pymupdf.pdfcolor["navy"])  # type: ignore : missing typing in pymupdf

            PdfPlotter.show_page(page, dpi=dpi)

    @staticmethod
    def _get_items_per_page(
        items_list: list[TextSpan | TextLine | TextBlock | PdfTable],
    ) -> dict[int, list[TextSpan | TextLine | TextBlock | PdfTable]]:
        """Groups the list of items per page. Takes as input
        any list of objects that have a "page" attribute
        Return a dict of structure is {page_number : sublist_of_items, ...}

        Args:
            items_list (list[TextSpan  |  TextLine  |  TextBlock]): A list
                of items to sort by page.

        Returns:
            dict[int,list]: The map of pages to sublists of items
        """
        return {
            page: list(items_on_page)
            for page, items_on_page in groupby(items_list, key=lambda item: item.page)
        }

    @staticmethod
    def show_page(page: pymupdf.Page, dpi: int = 100) -> None:
        """
        Generates an RGB Pixmap from the page and its rectangles,
        Args:
            page: a Page object of the document with rectangles drawn on it.
            dpi: the resolution of the generated image
        """
        plt.rcParams.update({"font.size": 5})
        pix = page.get_pixmap(dpi=dpi)  # type: ignore : missing typing in pymupdf | Page.get_pixmap(*args) -> Pixmap
        img = np.ndarray([int(pix.h), int(pix.w), 3], dtype=np.uint8, buffer=pix.samples_mv)  # type: ignore : missing typing in pymupdf | Pixmap.h : int ,Pixmap.mv : memoryview
        plt.figure(dpi=dpi)  # type: ignore : missing typing in matplotlib | plt.Figure(*args) -> Figure
        _ = plt.imshow(img, extent=(0, pix.w * 72 / dpi, pix.h * 72 / dpi, 0))  # type: ignore : missing typing in matplotlib | plt.imshow(*args) -> AxesImage

    @staticmethod
    def _get_rect_style(item: TextSpan) -> dict[str, tuple[int, int, int]]:
        """Get the kwargs for styling the rectangle, such as
        edge color and face color, considering the item's attributes

        Args:
            item (TextSpan): the item to get the styling params for

        Returns:
            dict: kwargs to be passed to the Shape.draw_rect() method
        """
        if item.is_header_footer:
            return {
                "color": pymupdf.pdfcolor["red"],  # type: ignore : missing typing in pymupdf | pymupdf.pdfcolor : dict[str | tuple[int, int, int]]
                "fill": pymupdf.pdfcolor["tomato"],  # type: ignore : missing typing in pymupdf | pymupdf.pdfcolor : dict[str | tuple[int, int, int]]
            }
        elif item.isin_table:
            return {
                "color": pymupdf.pdfcolor["steelblue"],  # type: ignore : missing typing in pymupdf | pymupdf.pdfcolor : dict[str | tuple[int, int, int]]
                "fill": pymupdf.pdfcolor["lightblue"],  # type: ignore : missing typing in pymupdf | pymupdf.pdfcolor : dict[str | tuple[int, int, int]]
            }
        else:
            return {
                "color": pymupdf.pdfcolor["goldenrod"],  # type: ignore : missing typing in pymupdf | pymupdf.pdfcolor : dict[str | tuple[int, int, int]]
                "fill": pymupdf.pdfcolor["gold"],  # type: ignore : missing typing in pymupdf | pymupdf.pdfcolor : dict[str | tuple[int, int, int]]
            }



from ..chunkers.tools.tools import Chunk
from .abstract_pipeline import AbstractPipeline


class BasePipeline(AbstractPipeline):

    def chunk_string(self, string: str) -> list[Chunk]:
        """Parses and chunks a string based on the provided
        parser and chunker.

        Args:
            string (str): a string.

        Returns:
            list[Chunk]: A list of chunks.
        """
        parsed_string = self.parser.parse_string(string)
        return self.chunker.chunk_string(parsed_string)

    def chunk_file(self, filepath: str) -> list[Chunk]:
        """Parses and chunks a string based on the provided
        parser and chunker.

        Args:
            filepath (Filepath): the path to a file.

        Returns:
            list[Chunk]: A list of chunks.
        """
        parsed_string = self.parser.parse_file(filepath)
        return self.chunker.chunk_string(parsed_string)



from collections import defaultdict
from operator import attrgetter
from typing import Literal
import pandas as pd
from pydantic import BaseModel, Field
import pymupdf  # type: ignore : no stubs

# TODO : Find list of special bullet-list caracters
INVALID_CHAR_MAP = {
    "\xa0": " ",
    "\uf0a7": "",
    "\uf0b7": "- ",
    "–": "- ",
    "•": "- ",
    "●": "- ",
    "►": "- ",
    "": "- ",  # arrow-like bullet
    "": "- ",  # triangle-like bullet
}


class Link:
    def __init__(self, uri: str, bbox: pymupdf.Rect) -> None:
        self.uri = uri
        self.bbox = bbox

    def __str__(self) -> str:
        return self.uri


class TocTitle(BaseModel):
    text: str = Field(description="The title's text.")
    source: Literal["metadata", "regex", "fontsize"] = (
        Field(description="The method used to obtain the title."),
    )
    page: int = (Field(description="The destination page refered by the title."),)
    level: int | None = Field(
        default=None,
        description="The title's level. From 1 to 6, such as html headers.",
        gt=0,
    )
    x_offset: float | None = Field(
        default=None,
        description="The horizontal offset from the left of the page of the title's span in the ToC section. Used only if title obtained from 'regex'.",
        gt=0,
    )
    source_page: int | None = Field(
        default=None,
        description="The page where the regex have been matched (supposedly the ToC section). Used only if title obtained from 'regex'.",
    )
    found: bool = Field(
        default=False,
        description="Whether or not the title has been found on the destination page.",
    )


class TextSpan:
    _bbox: pymupdf.Rect
    text: str
    font: str
    fontcolor: int
    fontsize: float
    flags: int
    ascender: float
    descender: float
    origin = pymupdf.Point
    page: int
    # attributes to be modified while processing the pdf
    order: int  # the order of the span among all spans
    isin_table: bool  # whether or not the span is in a table
    is_header_footer: bool  # whether the span is a header or footer
    link: Link | None  # The link bound to this span, if a link is bound

    def __init__(
        self,
        *,
        bbox: tuple[float],
        text: str,
        font: str,
        color: int,
        size: float,
        flags: int,
        ascender: float,
        descender: float,
        origin: tuple[float],
        page: int,
    ) -> None:
        self._bbox = pymupdf.Rect(bbox)
        self.text = TextSpan._remove_invalid_characters(text)
        self.font = font
        self.fontcolor = color
        self.fontsize = size
        self.flags = flags
        self.ascender = ascender
        self.descender = descender
        self.origin = pymupdf.Point(origin)
        self.page = page

        self.order = 0
        self.isin_table = False
        self.is_header_footer = False
        self.link = None

    @property
    def bbox(self) -> pymupdf.Rect:
        return self._bbox

    @property
    def is_superscripted(self) -> bool:
        return self.flags & 2**0

    @property
    def is_italic(self) -> bool:
        return self.flags & 2**1

    @property
    def is_serifed(self) -> bool:
        return self.flags & 2**2

    @property
    def is_monospaced(self) -> bool:
        return self.flags & 2**3

    @property
    def is_bold(self) -> bool:
        return self.flags & 2**4

    @property
    def line_height(self) -> float:
        return self.bbox.y1 - self.bbox.y0  # type: ignore : missing typing in pymuPdf | Rect.y0 : float, Rect.y1 : float

    @property
    def rgb_fontcolor(self) -> tuple[int, int, int]:
        return pymupdf.sRGB_to_rgb(self.fontcolor)  # type: ignore : missing typing in pymuPdf | pymupdf.sRGB_to_rgb(srgb: int) -> tuple[int, int, int]

    @property
    def is_empty(self) -> bool:
        return self.text.isspace()

    @staticmethod
    def _remove_invalid_characters(text: str) -> str:
        """Remove the invalid characters in a string

        Args:
            text (str): the text to cleaup

        Returns:
            str: the cleanedup text
        """
        # TODO : find more computationaly efficient method ? Maybe use str.translate() in conjunction with str.maketrans()
        for char, replacement in INVALID_CHAR_MAP.items():
            text = text.replace(char, replacement)

        return text

    def to_markdown(self) -> str:
        """Format the span's text to markdown format

        Returns:
            str: the markdown formatted text
        """
        if not self.text:
            return ""
        text = self.text.strip() if not self.link else f"[{self.text}]({self.link.uri})"
        if self.is_superscripted:
            text = f"<sup>{text}</sup>"
        if self.is_bold:
            text = f" **{text}** "
        if self.is_italic:
            text = f" *{text}* "

        return text

    def __str__(self) -> str:
        return self.text


class TextLine:
    """A line is a group of spans.

    Most of its attributes values reflect the attributes of most caracters in spans.
    Example : If a TextLine has 2 spans
        - TextSpan(text="blip", line_height=10)
        - TextSpan(text="blipbliblip", line_height=20)
        => then the line height if this textline will be 20, as most caracters have this feature.
    """

    spans: list[TextSpan]
    # attributes to be modified while parsing the document
    is_toc_element: bool

    def __init__(self, spans: list[TextSpan]) -> None:
        self.spans = spans

        self.is_toc_element = False

    @property
    def text(self) -> str:
        return "".join((span.text for span in self.spans))

    @property
    def bbox(self) -> pymupdf.Rect:
        bbox = self.spans[0].bbox
        for span in self.spans[1:]:
            bbox = bbox.include_rect(span.bbox)  # type: ignore : missing typing in pymuPdf | Rect.include_rect(r: Rect) -> Rect
        return bbox

    @property
    def line_height(self) -> float:
        lineheight_occurences_map: dict[float, int] = defaultdict(int)
        for span in self.spans:
            lineheight_occurences_map[span.line_height] += len(span.text)
        return max(lineheight_occurences_map, key=lineheight_occurences_map.get)

    @property
    def fontsize(self) -> float:
        fontsize_occurences_map: dict[float, int] = defaultdict(int)
        for span in self.spans:
            fontsize_occurences_map[span.fontsize] += len(span.text)
        return max(fontsize_occurences_map, key=fontsize_occurences_map.get)

    @property
    def origin(self) -> pymupdf.Point:
        return pymupdf.Point(
            min((span.origin.x for span in self.spans)),  # type: ignore : missing typing in pymuPdf | Point.x -> float
            max((span.origin.y for span in self.spans)),  # type: ignore : missing typing in pymuPdf | Point.y -> float
        )

    @property
    def page(self) -> int:
        return self.spans[0].page

    @property
    def is_empty(self) -> bool:
        return all(span.is_empty for span in self.spans)

    @property
    def is_bullet_point(self) -> bool:
        return self.text.startswith("- ")

    @property
    def order(self) -> int:
        return min(self.spans, key=lambda x: x.order).order

    def to_markdown(self) -> str:
        """Get the text content of all spans in the line, as markdown

        Returns:
            str: the markdown formatted text
        """
        return " ".join((span.to_markdown() for span in self.spans))

    def __str__(self) -> str:
        return self.text


class TextBlock:
    """A TextBlock is a list of lines"""

    lines: list[TextLine]

    # Attributes to modify while parsing the pdf
    section_title: TocTitle | None  # store the TocTitle if block is a ssection title

    def __init__(self, lines: list[TextLine]) -> None:
        self.lines = lines

        # Attributes to modify while parsing the pdf
        self.section_title = None  # store the TocTitle if block is a ssection title

    @property
    def text(self):
        return "".join((line.text for line in self.lines))

    @property
    def bbox(self) -> pymupdf.Rect:
        bbox = self.lines[0].bbox
        for line in self.lines[1:]:
            bbox = bbox.include_rect(line.bbox)  # type: ignore : missing typing in pymuPdf | Rect.include_rect(r: Rect) -> Rect
        return bbox

    @property
    def page(self) -> int:
        return self.lines[0].page

    @property
    def order(self) -> int:
        return min(self.lines, key=lambda x: x.order).order

    @property
    def is_empty(self) -> bool:
        return all(line.is_empty for line in self.lines)

    @property
    def fontsize(self) -> float:
        fontsize_occurences_map: dict[float, int] = defaultdict(int)
        for line in self.lines:
            fontsize_occurences_map[line.fontsize] += len(line.text)
        return max(fontsize_occurences_map, key=fontsize_occurences_map.get)

    def to_markdown(self) -> str:
        """Get the text content of all spans in the line, as markdown

        Returns:
            str: the markdown formated text
        """
        if self.is_empty:
            return ""
        md_string = ""
        for line in self.lines:
            if line.is_bullet_point or line.is_toc_element:
                md_string += "\n\n" + line.to_markdown().strip()
            else:
                md_string += " " + line.to_markdown().strip()
        if self.section_title:
            level = self.section_title.level or 0
            return "\n\n#" + "#" * level + " " + md_string.strip() + "\n\n"
        return md_string + "\n\n"

    def __str__(self) -> str:
        return self.text


class Cell(pymupdf.Rect):
    """A cell object. It inherit
    the pymupdf.Rect class but has PdfParser's spans
    bound to it
    """

    spans: list[TextSpan]

    def __init__(self, x0: float, y0: float, x1: float, y1: float) -> None:
        super().__init__(x0, y0, x1, y1)  # type: ignore : missing typing in pymuPdf : Rect.x0, Rect.y0, Rect.x1, Rect.y1 are all float
        # State
        self.spans = []


class PdfTable:
    """This class is intended to be used to parse the table of the document
    Note that PyMuPdf already implements a "Table" class.
    Though it has several issues :
    - wrongly parsed cell text (artifacts)
    - bad handling of merged cells
    - not fitting well with the mechanic of the PdfParser
    """

    cells: list[Cell]
    page: int

    def __init__(self, cells: list[Cell], page: int) -> None:
        """Initialize Table

        Args:
            cells (list[Cell]): a list of Cell object that are part of the table
            page (int): the page number the table belongs to
        """
        self.cells = cells
        self.page = page

    @property
    def bbox(self) -> pymupdf.Rect:
        return pymupdf.Rect(
            min(self.cells, key=lambda cell: cell.x0).x0,  # type: ignore : missing typing in pymuPdf | Rect.x0 : float
            min(self.cells, key=lambda cell: cell.y0).y0,  # type: ignore : missing typing in pymuPdf | Rect.y0 : float
            max(self.cells, key=lambda cell: cell.x1).x1,  # type: ignore : missing typing in pymuPdf | Rect.x1 : float
            max(self.cells, key=lambda cell: cell.y1).y1,  # type: ignore : missing typing in pymuPdf | Rect.y1 : float
        )

    @property
    def order(self) -> int:
        return min(
            (span for cell in self.cells for span in cell.spans),
            key=attrgetter("order"),
        ).order

    def get_table_grid(self) -> list[list[Cell]]:
        """Gets the grid of the table.
        As some self.cells might be merged cells, this
        method get the grid of cells if none wer merged

        Returns:
            list[list[Cells]] : the list of cells, organized by row
        """
        y_lines = sorted(
            list(
                set(cell.y0 for cell in self.cells)  # type: ignore : missing typing in pymuPdf | Rect.y0 : float
                | set(cell.y1 for cell in self.cells)  # type: ignore : missing typing in pymuPdf | Rect.y1 : float
            )
        )
        x_lines = sorted(
            list(
                set(cell.x0 for cell in self.cells)  # type: ignore : missing typing in pymuPdf | Rect.x0 : float
                | set(cell.x1 for cell in self.cells)  # type: ignore : missing typing in pymuPdf | Rect.x1 : float
            )
        )
        grid_cells: list[list[Cell]] = []
        for y0, y1 in zip(y_lines[:-1], y_lines[1:]):
            row_cells: list[Cell] = []
            for x0, x1 in zip(x_lines[:-1], x_lines[1:]):
                row_cells.append(Cell(x0, y0, x1, y1))
            grid_cells.append(row_cells)

        return grid_cells

    def to_pandas(self) -> pd.DataFrame:
        """Transforms this table to a pd.DataFrame

        Returns:
            pd.DataFrame : the table as dataframe
        """
        grid_cells = self.get_table_grid()
        df_constructor: list[list[str]] = []
        for grid_row in grid_cells:
            row_text: list[str] = []
            for grid_cell in grid_row:
                cell_text = ""
                for cell in self.cells:
                    if cell.contains(grid_cell):  # type: ignore : missing typing in pymuPdf | Rect.contains(r: Rect | Point) -> bool
                        for span in cell.spans:
                            cell_text += " " + span.text
                row_text.append(cell_text)
            df_constructor.append(row_text)

        df = pd.DataFrame(df_constructor[1:], columns=df_constructor[0])

        return df

    def to_markdown(self) -> str:
        """Builds a markdown string from the provided table

        Returns:
            str: the markdown string
        """
        return self.to_pandas().to_markdown(index=False)

    def __str__(self) -> str:
        return self.to_markdown()



import json
from pathlib import Path

from pydantic import validate_call, ValidationError

from ..abstract_parser import AbstractParser
from ...parsers.markdown.markdown_parser import MarkdownParser
from ...parsers.html.html_parser import HTMLParser
from ...schemas.schemas import WikitJSONDocument
from ...types.types import MarkdownString


class WikitJsonParser(AbstractParser):
    """This parser is intended to be used on
    the JSON files that are formatted according
    to the WikitJSONDocument interface.
    """

    @validate_call
    def parse_string(self, string: str) -> MarkdownString:
        """Parses a json string.

        Args:
            string (str): a json string.

        Returns:
            MarkdownString : the formatted string.
        """
        try:
            content = json.loads(string)
            content = WikitJSONDocument(**content)
        except (json.JSONDecodeError, ValidationError) as e:
            raise e

        return self.parse_wikit_json_document(content)

    @validate_call
    def parse_file(self, filepath: str) -> MarkdownString:
        """Parses a json file formatted according to Wikit's schema.

        Args:
            filepath (FilePath): the path to a file.

        Returns:
            MarkdownString: the formatted string.
        """
        content = self.read_file(filepath)

        return self.parse_wikit_json_document(content)

    def parse_wikit_json_document(self, content: WikitJSONDocument) -> MarkdownString:
        """Parses the has_part key json document
        according to the specified text format.

        Args:
            content (WikitJSONDocument): the content to be parsed.
        """
        has_part = content.has_part[0].text
        match content.file_format:
            case None:
                raise ValueError(
                    "No value under 'file_format' key. File format must be specified."
                )
            case "text/markdown":
                parser = MarkdownParser()
            case "text/html":
                parser = HTMLParser()
            case other:
                raise ValueError(
                    f"Only text/markdown and text/html are handle by WikitJsonParser. Got {other}."
                )

        return parser.parse_string(has_part)

    def read_file(self, filepath: str) -> WikitJSONDocument:
        """Reads a Markdown file

        Args:
            filepath (str): the path to the markdown file

        Returns:
            str: the markdown string
        """
        path = Path(filepath)
        if path.suffix != ".json":
            raise ValueError("Only .json files can be passed to WikitJsonParser.")
        with path.open("r", encoding="utf8") as file:
            content = json.load(file)
        try:
            content = WikitJSONDocument(**content)
        except ValidationError as e:
            raise e

        return content



import os
import re
from operator import attrgetter
from itertools import groupby

from ....types.types import MarkdownString
from .utils import PdfParserState


class PdfExport(PdfParserState):
    """This class is intended to be inherited by the PdfParser.
    It groups all functions related to exporting the Pdf based
    on the elements parsed by the PdfParser. Uses attributes
    such as self.spans, self.lines and self.blocks
    """

    def to_markdown(
        self,
        keep_track_of_page: bool = False,
    ) -> MarkdownString | dict[int, MarkdownString]:
        """Export parsed content to markdown.

        Args:
            keep_track_of_page (bool) : if True, returns a dict with
                dict[page_number : MarkdownString]. Page number is 1-based
                and page 0 is reserved for main document title.
                If False returns a string for the whole document.
                Defaults to False.

        Returns:
            str | list[str]: the markdown string of the specified page range.
        """
        blocks_by_page = {
            page: list(blocks)
            for page, blocks in groupby(self.blocks, key=attrgetter("page"))
        }
        tables_by_page = {
            page: list(blocks)
            for page, blocks in groupby(self.tables, key=attrgetter("page"))
        }

        prefix = f"# {self.main_title}\n\n" if self.main_title else ""
        string_per_page = {}

        for page_n in range(self.page_start, self.page_end):
            blocks_on_page = blocks_by_page[page_n] if page_n in blocks_by_page else []
            tables_on_page = tables_by_page[page_n] if page_n in tables_by_page else []
            items_to_export = sorted(
                blocks_on_page + tables_on_page, key=attrgetter("order")
            )
            md_string = "\n\n".join(item.to_markdown() for item in items_to_export)
            md_string = PdfExport._cleanup_md_string(md_string)
            string_per_page[page_n + 1] = md_string.strip()

        if keep_track_of_page:
            return {
                page_n: MarkdownString(content=md_string)
                for page_n, md_string in string_per_page.items()
            }

        return MarkdownString(content=prefix + "\n\n".join(string_per_page.values()))

    @staticmethod
    def _cleanup_md_string(md_string: str) -> str:
        """Performs various cleanup operation
        of the markwdown string output by to_markdown()

        Args:
            md_string (str): the markdown string

        Returns:
            str : the cleaned up string
        """
        md_string = re.sub(r"\*\* *\*\*", " ", md_string)
        md_string = md_string.replace(" ** ", "**\n")
        md_string = re.sub(r"\n{3,}", "\n\n", md_string)

        return md_string

    def save_markdown(self, output_filepath: str = "./output.md") -> None:
        """Generates the markdown export of the pdf
        and saves it in a file

        Args:
            output_filpath (str): the file path where the md is saved
        """
        md_string = self.to_markdown()
        if not os.path.exists(os.path.dirname(output_filepath)):
            os.makedirs(os.path.dirname(output_filepath))
        with open(output_filepath, "w", encoding="utf-8") as f:
            f.write(md_string)

    def to_text(self) -> str:
        """
        Converts the parsed document to a text string
        without formatting

        Returns:
            str: the raw text string
        """
        raise NotImplementedError()



from inspect import signature
from typing import Any, Callable


def validate_args(function: Callable[..., Any]) -> Any:
    """Meant to be used as a decorator : @validate_args.
    Checks that the types of arguments passed to a function
    matches the typing provided.

    Note : pydantic proposes a similar decorator : @validate_call.
    HOWEVER, this causes an error when applied to __ini__() functions.

    Args:
        function (Callable[..., Any]): any typed function.

    Returns:
        Any: the return of the function.
    """

    def wrapper(*args: tuple[Any], **kwargs: dict[Any, Any]) -> Any:
        sig = signature(function)
        for arg_name, arg_value in zip(sig.parameters, args):
            if (
                sig.parameters[arg_name].annotation
                is not sig.parameters[arg_name].empty
            ):
                if not isinstance(arg_value, sig.parameters[arg_name].annotation):
                    raise ValueError(
                        f"Argument '{arg_name}' must be of type {sig.parameters[arg_name].annotation.__name__}."
                    )
        result = function(*args, **kwargs)
        return result

    return wrapper



from abc import ABC, abstractmethod

from pydantic import validate_call

from ..types.types import AbstractInOutType


class AbstractParser(ABC):
    """A parser is meant to parse either a string
    of a document provided as a file path.
    Either way, it is expected to return a formatted
    string belonging to one of the option specified
    in TypeString.
    IMPORTANT : The types specified in the 3 abstract methods
    must be the same.
    """

    @validate_call
    @abstractmethod
    def parse_string(self, string: str) -> AbstractInOutType:
        pass

    @validate_call
    @abstractmethod
    def parse_file(self, filepath: str) -> AbstractInOutType:
        pass



from typing import List, Literal
from pydantic import BaseModel, ConfigDict, Field


class DocumentJsonChunkedSchemaType(BaseModel):
    type: Literal["Document", "DocumentChunk"] = Field(
        ...,
        alias="@type",
        title="Type of the chunk",
        description="Type of the chunk, can be Document or DocumentChunk",
    )


class DocumentChunkMetadata(BaseModel):
    """Document Chunk common metadata"""

    page_start: int | None = Field(
        alias="pageStart",
        title="Start page in a pageable Document (e.g. a PDF file)",
        default=None,
    )
    page_end: int | None = Field(
        alias="pageEnd", title="End page in a pageable Document", default=None
    )
    position: int | None = Field(title="Position of the Document Chunk", default=None)


class WikitJSONDocumentChunk(DocumentJsonChunkedSchemaType, DocumentChunkMetadata):
    """
    Chunk of the document for the chunked JSON format upload
    """

    model_config = ConfigDict(populate_by_name=True)

    text: str = Field(title="Text of the chunk", default=None)

    title: str | None = Field(
        title="Title metadata",
        default=None,
    )


class WikitJSONDocument(DocumentJsonChunkedSchemaType):
    """
    Other fields of the document for the chunked JSON format upload
    """

    model_config = ConfigDict(populate_by_name=True)

    type: Literal["Document", "DocumentChunk"] = Field(
        ...,
        alias="@type",
        title="Type of the chunk",
        description="Type of the chunk, can be Document or DocumentChunk",
    )

    has_part: List[WikitJSONDocumentChunk] = Field(
        alias="hasPart", title="Chunks of the document", default=None
    )

    context: str | None = Field(..., alias="@context", title="Context of the document")

    identifier: str | None = Field(title="Identifier of the document", default=None)

    name: str | None = Field(title="Name of the document", default=None)

    title: str | None = Field(title="Title of the document", default=None)

    text: str | None = Field(title="Text of the document", default=None)

    description: str | None = Field(title="Description of the document", default=None)

    source: str | None = Field(title="Source of the document", default=None)

    language: str | None = Field(title="Language of the document", default=None)

    url: str | None = Field(title="Url of the document", default=None)

    keywords: List[str] | None = Field(title="Keywords of the document", default=None)

    comment: str | None = Field(title="Comment of the document", default=None)

    date_created: str | None = Field(
        alias="dateCreated", title="Date of creation", default=None
    )

    date_modified: str | None = Field(
        alias="dateModified", title="Date of the last modification", default=None
    )

    file_format: str | None = Field(
        alias="fileFormat", title="File format of the document", default=None
    )



from ..decorators.decorators import validate_args
from ..types.types import MarkdownString
from ..chunkers.markdown_chunker import MarkdownChunker
from ..chunkers.tools.tools import Chunk
from ..parsers.pdf.pdf_parser import PdfParser


class PdfPipeline:
    """This pipeline is meant to be used on for pdf files.
    First, it uses a parser to parse the pdf to markdown format.
    Second, it build chunks based on headers if any have been found,
    or pages if not.
    """

    parser: PdfParser
    chunker: MarkdownChunker

    @validate_args
    def __init__(self, parser: PdfParser, chunker: MarkdownChunker) -> None:
        self.parser = parser
        self.chunker = chunker

    def chunk_string(
        self, string: bytes, page_start: int = 0, page_end: int | None = None
    ) -> list[Chunk]:
        """
        Chunks a PDF byte stream.
        It first convert the string to markdown using the PdfParser,
        and then chunks the markdown file.

        Args:
            filepath (str): the path to the pdf file to chunk
            page_start (int, optional): the page to start parsing from. Defaults to 0.
            page_end (int, optional): the page to stop parsing. None to parse until last page. Defaults to None.

        Returns:
            Chunks: the chunks
        """
        _ = self.parser.parse_string(string, page_start, page_end)
        return self._get_chunks_using_strategy()

    def chunk_file(
        self, filepath: str, page_start: int = 0, page_end: int | None = None
    ) -> list[Chunk]:
        """
        Chunks a PDF file.
        It first convert the pdf file to markdown using the PdfParser,
        and then chunks the markdown file.

        Args:
            filepath (str): the path to the pdf file to chunk
            page_start (int, optional): the page to start parsing from. Defaults to 0.
            page_end (int, optional): the page to stop parsing. None to parse until last page. Defaults to None.

        Returns:
            Chunks: the chunks
        """
        _ = self.parser.parse_file(filepath, page_start, page_end)
        return self._get_chunks_using_strategy()

    def _get_chunks_using_strategy(self) -> list[Chunk]:
        """Handles the routing toward the adequate chunking
        strategy.
        The chunking strategy is as follow :
        - if table of content found in document -> chunk based on table of content
        - if not -> chunk by page

        Returns:
            Chunk: the list of chunks
        """
        doc_orientation = self.parser.get_document_orientation()

        if self.headers_have_been_found():
            chunks = self.chunk_with_headers()
        elif doc_orientation == "landscape":
            chunks = self.chunk_by_page()
        # placeholder : if no headers found and document is portrait, then chunk with headers
        # as no header have been found, this will result in on big chunk being chunk into subchunks by word length
        # In the future, a method will be implemented to detect TOC using advanced techniques.
        else:
            chunks = self.chunk_with_headers()

        self.parser.cleanup_memory()

        return chunks

    def headers_have_been_found(self) -> bool:
        """Determines whether or not enough header have been
        found in order to chunk document using MarkdownChunker.
        If more than half the headers have been found, returns True.

        Returns:
            bool : True if detected headers have been found in document.
        """
        if not self.parser.toc:
            return False
        headers_found = [header for header in self.parser.toc if header.found]
        return len(headers_found) / len(self.parser.toc) > 0.5

    def chunk_with_headers(self) -> list[Chunk]:
        """Uses the MarkdownChunker to chunk the document
        based on headers.

        Returns:
            Chunks: the list of chunks
        """
        strings_per_page: dict[int, MarkdownString] = self.parser.to_markdown(
            keep_track_of_page=True
        )
        main_title = f"# {self.parser.main_title}\n\n" if self.parser.main_title else ""
        md_string = MarkdownString(
            content=main_title
            + "\n\n".join((string.content for string in strings_per_page.values()))
        )
        chunks = self.chunker.chunk_string(md_string)
        # build map dict[index of line : page_number] to keep track of the line index for each page
        # Add start line attribute
        line_to_page_map = {}
        line_idx = len(main_title.split("\n"))
        for page_n in sorted(strings_per_page.keys()):
            line_to_page_map[page_n] = line_idx
            line_idx += len(strings_per_page[page_n].content.split("\n"))
        for chunk in chunks:
            for page_n, line_idx in line_to_page_map.items():
                if line_idx < chunk.start_line:
                    chunk.start_page = page_n
        chunks[0].start_page = min(line_to_page_map.keys())
        # Add end line attribute
        for chunk, next_chunk in zip(chunks[:-1], chunks[1:]):
            chunk.end_page = next_chunk.start_page
        chunks[-1].end_page = list(strings_per_page.keys())[-1]

        return chunks

    def chunk_by_page(self) -> list[Chunk]:
        """Build one chunk per page.

        Returns:
            Chunks: the list of chunks.
        """
        strings_per_page: dict[int, MarkdownString] = self.parser.to_markdown(
            keep_track_of_page=True
        )
        main_title = f"# {self.parser.main_title}\n\n" if self.parser.main_title else ""

        chunks: list[Chunk] = []
        line_idx = 0
        for page_n, md_string in strings_per_page.items():
            chunks.append(
                Chunk(
                    headers=[main_title],
                    content=md_string.content,
                    start_page=page_n,
                    end_page=page_n + 1,
                    start_line=line_idx,
                )
            )
            line_idx += len(md_string.content.split("\n"))

        chunks = self.chunker.split_big_chunks(chunks)

        return chunks

    def save_chunks(
        self, chunks: list[Chunk], output_filename: str, remove_links: bool = False
    ) -> None:
        """Saves the chunks at the designated location
        as a json list of chunks.

        Args:
            chunks (Chunks): the chunks.
            output_filename (str): the JSON file where to save the files. Must be json.
            remove_links (bool): Whether or not links should be remove from the chunk's text content.
        """
        self.chunker.save_chunks(chunks, output_filename, remove_links)



import re
from typing import Literal

from .abstract_chunker import AbstractChunker
from .tools.tools import Chunk, TocTree
from ..types.types import MarkdownString


class MarkdownChunker(AbstractChunker):
    max_headers_to_use: str
    max_chunk_word_count: int
    hard_max_chunk_word_count: int
    min_chunk_word_count: int

    def __init__(
        self,
        max_headers_to_use: Literal["h1", "h2", "h3", "h4", "h5", "h6"] = "h4",
        max_chunk_word_count: int = 200,
        hard_max_chunk_word_count: int = 400,
        min_chunk_word_count: int = 15,
    ) -> None:
        """Initialize a Markdown chunker

        Args:
            max_headers_to_use (MaxHeadersToUse) : The maximum header level to consider (included).
                Headers with level lower than this wont be used to split chunks.
                For example, if 'h4' is set, then 'h5' and 'h6' headers won't be used.
                Must be a string of type 'hx' with x being the title level ranging from 1 to 6.
            max_chunk_word_count (int) : The maximum size a chunk can be (in words).
                It is a SOFT limit, meaning that chunks bigger that this size will be chunked using lower level headers if any are available."
            hard_max_chunk_word_count (int) : The true maximum size a chunk can be (in word).
                It is a HARD limit, meaning that chunks bigger by this limit will be split into subchunks.
                ChunkNorris will try to equilibrate the size of resulting subchunks. It should be greater than max_chunk_word_count."
            min_chunk_word_count (int) : The minimum size a chunk can be (in words).
                Chunks lower than this will be discarded.
        """
        self.max_headers_to_use = max_headers_to_use
        self.max_chunk_word_count = max_chunk_word_count
        self.hard_max_chunk_word_count = hard_max_chunk_word_count
        self.min_chunk_word_count = min_chunk_word_count

    @property
    def _header_regex_patterns(self) -> dict[str, re.Pattern[str]]:
        """The header regex patterns for headers."""
        patterns = {
            "h1": re.compile(r"^(?:- )?(#{1} .+)", re.MULTILINE),
            "h2": re.compile(r"^(?:- )?(#{2} .+)", re.MULTILINE),
            "h3": re.compile(r"^(?:- )?(#{3} .+)", re.MULTILINE),
            "h4": re.compile(r"^(?:- )?(#{4} .+)", re.MULTILINE),
            "h5": re.compile(r"^(?:- )?(#{5} .+)", re.MULTILINE),
        }

        return patterns

    def chunk_string(self, string: MarkdownString) -> list[Chunk]:
        """Chunks a Markdown formatted string

        Returns:
            Chunks: the chunks
        """
        toc_tree = self.get_toc_tree(string.content)
        chunks = self.get_chunks(toc_tree)

        return chunks

    def get_toc_tree(
        self,
        md_string: str,
    ) -> TocTree:
        """Builds the table of content tree based on header

        Args:
            md_string (str): the markdown formatted string

        Returns:
            TocTree: the table of content
        """
        headers_patterns_to_use = [
            self._header_regex_patterns[f"h{i}"]
            for i in range(1, int(self.max_headers_to_use[1]) + 1)
        ]
        tree = TocTree()
        current_node = tree
        id_cntr = 0
        match = None
        within_code_block = False
        for line_idx, line in enumerate(md_string.split("\n")):
            if line.startswith("```") and not "```" in line[3:]:
                within_code_block = not within_code_block
            if not within_code_block:
                for level, pattern in enumerate(headers_patterns_to_use):
                    match = re.match(pattern, line)
                    if match and not within_code_block:
                        title = match.group(1)
                        while (
                            current_node.parent is not None
                            and level <= current_node.level
                        ):
                            current_node = current_node.parent
                        current_node.add_child(
                            TocTree(
                                title=title,
                                id=id_cntr,
                                line_index=line_idx,
                                level=level,
                            )
                        )
                        current_node = current_node.children[-1]
                        id_cntr += 1
                        break
            if within_code_block or not match:
                current_node.content += line + "\n"

        return tree

    def get_chunks(self, toc_tree: TocTree) -> list[Chunk]:
        """Wrapper that build the chunk's texts, check
        that they fit in size, replace links formatting.

        Args:
            toc_tree (TocTree): the toc tree of the document

        Returns:
            Chunks: the chunks text, formatted
        """
        chunks = self.build_chunks(toc_tree)
        chunks = self.remove_small_chunks(chunks)
        chunks = self.split_big_chunks(chunks)

        return chunks

    def build_chunks(
        self,
        toc_tree_element: TocTree,
        already_ok_chunks: list[Chunk] | None = None,
    ) -> list[Chunk]:
        """Uses the toc tree to build the chunks. Uses recursion.
        Method :
        - build the chunk (= titles from sections above + section content + content of subsections)
        - if the chunk is too big:
            - save the section as title + content (if section has content)
            - subdivide section recursively using subsections
        - else save it as is

        Args:
            toc_tree_element (TocTree): the TocTree for which the chunk should be build
            already_ok_chunks (Chunks, optional): the chunks already built.
                Used for recursion. Defaults to None.

        Returns:
            Chunks: list of chunk's texts
        """
        if not already_ok_chunks:
            already_ok_chunks = []

        current_chunk = MarkdownChunker._build_chunk(toc_tree_element)

        if current_chunk.word_count > self.max_chunk_word_count:
            if toc_tree_element.content.strip():
                parent_headers = MarkdownChunker.get_parents_headers(toc_tree_element)
                current_chunk = Chunk(
                    headers=parent_headers + [toc_tree_element.title],
                    content=toc_tree_element.content,
                    start_line=toc_tree_element.line_index,
                )
                already_ok_chunks.append(current_chunk)
            for child in toc_tree_element.children:
                already_ok_chunks = self.build_chunks(child, already_ok_chunks)
        else:
            already_ok_chunks.append(current_chunk)

        return already_ok_chunks

    @staticmethod
    def _build_chunk(toc_tree_element: TocTree) -> Chunk:
        """Builds a chunk by apposing the text of headers
        and recursively getting the content of children

        Args:
            toc_tree_element (TocTree): the toc tree element

        Returns:
            str: the chunk content. parent's headers + content
        """
        parent_headers = MarkdownChunker.get_parents_headers(toc_tree_element)
        content = MarkdownChunker._build_chunk_content(toc_tree_element)

        return Chunk(
            headers=parent_headers,
            content=content,
            start_line=toc_tree_element.line_index,
        )

    @staticmethod
    def _build_chunk_content(toc_tree_element: TocTree) -> str:
        """Builds a chunk content (i.e without headers above)
        from a toc tree. It uses the toc tree's content, and recursively
        adds the header and content of its children

        Args:
            toc_tree_element (TocTree): the toc tree (or element of toc tree)

        Returns:
            str: the text of the chunk (without the headers of parents)
        """
        text = toc_tree_element.title + "\n\n" + toc_tree_element.content
        for child in toc_tree_element.children:
            text += "\n\n" + MarkdownChunker._build_chunk_content(child)

        return text

    @staticmethod
    def get_parents_headers(toc_tree_element: TocTree) -> list[str]:
        """Gets a list of the titles that are parent
        of the provided toc tree element. The list
        is ordered in descending order in terms of header level.

        Args:
            toc_tree_element (TocTree): the toc tree element

        Returns:
            list[str]: the list of header's text
        """
        headers: list[str] = []
        while toc_tree_element.parent:
            toc_tree_element = toc_tree_element.parent
            headers.append(toc_tree_element.title)
        # remove empty string header, such as root header
        headers = [h for h in headers if h]

        return list(reversed(headers))

    def remove_small_chunks(self, chunks: list[Chunk]) -> list[Chunk]:
        """Removes chunks that have less words than the specified limit

        Args:
            chunks (Chunks): the list of chunks

        Returns:
            Chunks: the chunks with more words than the specified threshold
        """
        return [c for c in chunks if c.word_count >= self.min_chunk_word_count]

    def split_big_chunks(
        self,
        chunks: list[Chunk],
    ) -> list[Chunk]:
        """Splits the chunks that are too big.
        You may consider passing the kwarg "hard_max_chunk_word_count"
        to specify the limit size of the chunk (in words)

        Args:
            chunks (Chunks): The chunks obtained from the get_chunks() method

        Returns:
            Chunks: the chunks, with big chunks splitting into smaller chunks
        """
        splitted_chunks: list[Chunk] = []
        for chunk in chunks:
            if chunk.word_count < self.hard_max_chunk_word_count:
                splitted_chunks.append(chunk)
            else:
                splitted_chunk = self._split_with_newlines(chunk)
                splitted_chunks.extend(splitted_chunk)

        return splitted_chunks

    def _split_with_newlines(
        self,
        chunk: Chunk,
    ) -> list[Chunk]:
        """Split chunks based on newlines. Adds the
        chunk titles at the beginning of each chunk

        Args:
            chunk (Chunk): the chunk to split
        """
        split_count = (chunk.word_count // self.hard_max_chunk_word_count) + 1
        split_word_size = chunk.word_count // split_count

        chunks: list[Chunk] = []
        current_chunk_lines: list[str] = []
        headers_lines: list[str | None] = [None] * 6  # to store the headers
        headers_lines_buffer = headers_lines.copy()
        prev_start_line = chunk.start_line
        within_code_block = False
        for line_idx, line in enumerate(chunk.content.split("\n")):
            if line.startswith("```") and not "```" in line[3:]:
                within_code_block = not within_code_block
            if line.startswith("#") and not within_code_block:  # header
                header_lvl = len(line) - len(line.lstrip("#")) - 1
                headers_lines[header_lvl:] = [None] * (len(headers_lines) - header_lvl)
                headers_lines[header_lvl] = line
            current_chunk_lines.append(line)
            if len("\n".join(current_chunk_lines).split()) > split_word_size:
                headers = list(filter(None, headers_lines_buffer))
                new_chunk = Chunk(
                    headers=chunk.headers,
                    content="\n\n".join(headers)
                    + "\n\n"
                    + "\n".join(current_chunk_lines),
                    start_line=prev_start_line,
                    start_page=chunk.start_page,
                    end_page=chunk.end_page,
                )
                chunks.append(new_chunk)
                prev_start_line = chunk.start_line + line_idx
                current_chunk_lines = []
                headers_lines_buffer = headers_lines.copy()
        if current_chunk_lines:
            headers = list(filter(None, headers_lines))
            new_chunk = Chunk(
                headers=chunk.headers,
                content="\n\n".join(headers) + "\n\n" + "\n".join(current_chunk_lines),
                start_line=prev_start_line,
                start_page=chunk.start_page,
                end_page=chunk.end_page,
            )
            chunks.append(new_chunk)

        return chunks



from pathlib import Path
import re
from markdownify import markdownify  # type: ignore : no stub file

from ...types.types import MarkdownString
from ..abstract_parser import AbstractParser


class HTMLParser(AbstractParser):

    def parse_string(self, string: str) -> MarkdownString:
        """Parses a markdown-formatted string.
        Ensures that the formatting is suited to be passed
        to the MarkdownChunker.

        Args:
            string (str): the markdown formatted string

        Returns:
            TypedString: the formatted markdown string
        """
        formatted_string = HTMLParser.apply_markdownify(string)
        formatted_string = HTMLParser.cleanup_string(formatted_string)

        return MarkdownString(
            content=formatted_string,
        )

    def parse_file(self, filepath: str) -> MarkdownString:
        """Reads and parses a markdown-formatted string.
        Ensures that the formatting is suited to be passed
        to the MarkdownChunker.

        Args:
            filepath (FilePath): the path to a .md file

        Returns:
            TypedString: the typed string
        """
        html_string = HTMLParser.read_file(filepath)

        return self.parse_string(html_string)

    @staticmethod
    def read_file(filepath: str) -> str:
        """Reads a Markdown file

        Args:
            filepath (str): the path to the markdown file

        Returns:
            str: the markdown string
        """
        path = Path(filepath)
        if path.suffix != ".html":
            raise ValueError("Only .html files can be passed to HTMLParser.")
        with path.open("r", encoding="utf8") as file:
            html_string = file.read()

        return html_string

    @staticmethod
    def apply_markdownify(html_string: str) -> str:
        """Applies markdownify to the html text

        Args:
            html_text (str): the text of the html file

        Returns:
            str: the markdownified string
        """
        md_string = markdownify(
            html_string,
            heading_style="ATX",
            strip=["figure", "img"],
            bullets="-*+",
            escape_asterisks=False,
            escape_underscores=False,
            escape_misc=False,
        )

        return md_string

    @staticmethod
    def cleanup_string(md_string: str) -> str:
        """Cleans up the html string,
        essentially by removing consecutive \n

        Args:
            html_string (str): the markdown string, output from
                apply_markdownify()

        Returns:
            str: the cleaned up string
        """
        md_string = md_string.strip()
        md_string = re.sub(r"\n{3,}", "\n\n", md_string)

        return md_string



from pathlib import Path
from ...types.types import MarkdownString
from ..abstract_parser import AbstractParser


class MarkdownParser(AbstractParser):

    def parse_string(self, string: str) -> MarkdownString:
        """Parses a markdown-formatted string.
        Ensures that the formatting is suited to be passed
        to the MarkdownChunker.

        Args:
            string (str): the markdown formatted string

        Returns:
            TypedString: the formatted markdown string
        """
        formatted_string = MarkdownParser.convert_setext_to_atx(string)

        return MarkdownString(
            content=formatted_string,
        )

    def parse_file(self, filepath: str) -> MarkdownString:
        """Reads and parses a markdown-formatted string.
        Ensures that the formatting is suited to be passed
        to the MarkdownChunker.

        Args:
            filepath (FilePath): the path to a .md file

        Returns:
            TypedString: the typed string
        """
        md_string = MarkdownParser.read_file(filepath)

        return self.parse_string(md_string)

    @staticmethod
    def read_file(filepath: str) -> str:
        """Reads a Markdown file

        Args:
            filepath (str): the path to the markdown file

        Returns:
            str: the markdown string
        """
        path = Path(filepath)
        if path.suffix != ".md":
            raise ValueError("Only .md files can be passed to MarkDownParser.")
        with path.open("r", encoding="utf8") as file:
            md_string = file.read()

        return md_string

    @staticmethod
    def convert_setext_to_atx(md_string: str) -> str:
        """Converts headers from setext style to atx style

        Args:
            md_string (str): the markdown string

        Return:
            str: the string with formatted headers
        """
        output_lines: list[str] = []
        within_code_block = False
        prev_line = ""

        for line in md_string.split("\n"):
            if line.startswith("```") and "```" not in line[3:]:
                within_code_block = not within_code_block
            if not within_code_block:
                if line.startswith("==="):
                    output_lines.pop()
                    output_lines.append(f"# {prev_line}")
                elif line.startswith("---"):
                    output_lines.pop()
                    output_lines.append(f"## {prev_line}")
                else:
                    output_lines.append(line)
            else:
                output_lines.append(line)
            prev_line = line

        return "\n".join(output_lines)



from .abstract_parser import AbstractParser
from .markdown.markdown_parser import MarkdownParser
from .html.html_parser import HTMLParser
from .json.wikit_parser import WikitJsonParser
from .pdf.pdf_parser import PdfParser



from __future__ import annotations
from unicodedata import normalize
from copy import deepcopy
import re
from typing import Optional
import json


class Chunk:
    headers: list[str]
    content: str
    start_line: int
    start_page: Optional[int]
    end_page: Optional[int]

    def __init__(
        self,
        headers: list[str],
        content: str,
        start_line: int,
        *,
        start_page: Optional[int] = None,
        end_page: Optional[int] = None,
    ) -> None:
        self.headers = headers
        self.content = Chunk._cleanup_text(content)
        self.start_line = start_line
        self.start_page = start_page
        self.end_page = end_page

    @property
    def word_count(self) -> int:
        return len(self.content.split())

    def __str__(self) -> str:
        return self.get_text()

    def get_text(self, remove_links: bool = False) -> str:
        """Gets the text of the chunk.

        Args:
            remove_links (bool, optional): If True, the markdown links will be removed (text of the link is kept).
                Defaults to False.

        Returns:
            str: the text
        """
        text = "\n\n".join(self.headers + [self.content])
        if remove_links:
            text = Chunk.remove_links(text)

        return Chunk._cleanup_text(text)

    @staticmethod
    def remove_links(text: str) -> str:
        """Removes the markdown format of the links in the text.

        Args:
            text (str): the text to find the links in

        Returns:
            str: the formated text
        """
        pattern = r"\[?!?\[(.*?)\](\(.*\)\])?\((.*\..+?)(\s.*)?\)"
        matches = re.finditer(pattern, text)
        for match in matches:
            text = text.replace(match[0], match[1])

        return text

    @staticmethod
    def _cleanup_text(text: str) -> str:
        """Cleans up the text"""
        text = normalize("NFKD", text)
        text = re.sub(r"\n{3,}", "\n\n", text)
        text = text.strip()

        return text


class TocTree:
    id: int
    text: str
    level: int
    line_index: int
    content: str
    parent: TocTree | None
    children: list[TocTree]

    def __init__(
        self,
        *,
        title: str = "",
        children: list[TocTree] | None = None,
        content: str = "",
        id: int = -1,
        line_index: int = -1,
        level: int = -1,
        parent: TocTree | None = None,
    ) -> None:
        self.title = title
        self.content = content
        self.id = id
        self.line_index = line_index
        self.level = level
        self.parent = parent
        self.children = [] if children is None else children

    def add_child(self, child: TocTree) -> None:
        """Adds a child to the list of TocTree.
        Self is added as parent

        Args:
            child (TocTree): a TocTree node to add to the children.
        """
        child.parent = self
        self.children.append(child)

    def get_title_by_id(self, id: int) -> TocTree | None:
        """Gets a toc tree using its id.

        Args:
            id (int): the id of the title we are looking for.

        Returns:
            TocTree: the element of title we were looking for
        """
        if self.id == id:
            return self
        for child in self.children:
            target = child.get_title_by_id(id)
            if target:
                return target

    def to_json(self, output_path: str = "./tree.json") -> None:
        """Outputs the tree as json.

        Returns:
            dict[str, Any]: the json
        """
        dumpable_tree = deepcopy(self)
        dumpable_tree.remove_circular_refs()
        with open(output_path, "w", encoding="utf-8") as file:
            json.dump(
                dumpable_tree,
                file,
                default=lambda o: o.__dict__,
                sort_keys=True,
                ensure_ascii=False,
            )

    def remove_circular_refs(self):
        """Recursively removes the circular ref in dict
        (used to save as json).

        Args:
            tree (TocTree): a toc tree to remove circular refs
        """
        self.parent = None
        for child in self.children:
            child.remove_circular_refs()



import glob
import os
import json
from argparse import ArgumentParser

from ..chunkers.markdown_chunker import MarkdownChunker
from ..chunkers.tools.tools import Chunk
from ..parsers.json.wikit_parser import WikitJsonParser
from ..schemas.schemas import WikitJSONDocument, WikitJSONDocumentChunk
from ..decorators.decorators import validate_args


class WikitJsonPipeline:
    parser: WikitJsonParser
    chunker: MarkdownChunker

    @validate_args
    def __init__(self, parser: WikitJsonParser, chunker: MarkdownChunker) -> None:
        self.parser = parser
        self.chunker = chunker

    def chunk_file(self, filepath: str) -> list[Chunk]:
        """Chunks a json file formatted
        according to wikit's schema.

        Args:
            filepath (str): the path to the json file.

        Returns:
            list[Chunk]: the list of chunks
        """
        string = self.parser.parse_file(filepath)
        return self.chunker.chunk_string(string)

    def chunk_string(self, string: str) -> list[Chunk]:
        """Chunks a json string formatted according to
        wikit's schema.

        Args:
            string (str): the json string, obtained
                by using json.dumps() for the file.

        Returns:
            list[Chunk]: the list of chunks
        """
        md_string = self.parser.parse_string(string)
        return self.chunker.chunk_string(md_string)

    def chunk_and_save(self, filepath: str, output_filepath: str | None = None) -> None:
        """Chunks the file and saves the chunks
        in the WikiJsonDocument format.

        Args:
            filepath (str): the path to the json file.
            output_filepath (str): the path of the output json.
        """
        if not output_filepath:
            output_filepath = filepath.replace(".json", "-chunked.json")
        json_content = self.parser.read_file(filepath)
        md_string = self.parser.parse_wikit_json_document(json_content)
        chunks = self.chunker.chunk_string(md_string)
        json_content.has_part = self._format_chunks(chunks)
        self._save_content_as_wikit_json(json_content, output_filepath)

    def chunk_directory(self, input_dir: str) -> None:
        """Chunks the json files of entire directory, recursively

        Args:
            input_dir (str): the path to directory
        """
        if not os.path.isdir(input_dir) and os.path.exists(input_dir):
            raise ValueError("input_dir must point to an existing directory.")

        output_dir = os.path.dirname(input_dir) + "-chunked"

        filepaths: list[str] = [
            subdir
            for dir in os.walk(input_dir)
            for subdir in glob.glob(os.path.join(dir[0], "*.json"))
        ]

        for filepath in filepaths:
            output_filepath = filepath.replace(input_dir, output_dir)
            self.chunk_and_save(filepath, output_filepath)

    def _save_content_as_wikit_json(
        self, content: WikitJSONDocument, out_filepath: str
    ) -> None:
        """Saves each chunk in a MarkDown file.

        Args:
            out_filepath (str): the path where to save the output
            json_content (dict): the content of the json to save.
        """
        output_dir = os.path.dirname(out_filepath)
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        with open(out_filepath, "w", encoding="utf-8") as file:
            json.dump(content.model_dump(), file, ensure_ascii=False)

    def _format_chunks(self, chunks: list[Chunk]) -> list[WikitJSONDocumentChunk]:
        """Formats the chunks according to the input json file
        i.e places the chunks inside the key [hasPart]

        Args:
            json_content (dict): the json content of the source json file. Obtained from
                read_file().
            chunks (Chunks): the chunks obtained.

        Returns:
            dict: the formatted file content
        """
        has_part = [
            WikitJSONDocumentChunk(type="DocumentChunk", text=chunk.get_text(), position=i)  # type: ignore : known problem where pydantic looks for @type instead of type. see https://github.com/pydantic/pydantic/issues/4936
            for i, chunk in enumerate(chunks)
        ]

        return has_part

    def save_chunks(
        self, chunks: list[Chunk], output_filename: str, remove_links: bool = False
    ) -> None:
        """Saves the chunks at the designated location
        as a json list of chunks.

        Args:
            chunks (Chunks): the chunks.
            output_filename (str): the JSON file where to save the files. Must be json.
            remove_links (bool): Whether or not links should be remove from the chunk's text content.
        """
        self.chunker.save_chunks(chunks, output_filename, remove_links)


def parse_arguments():
    """Parse the given command-line arguments."""

    parser = ArgumentParser(
        description="Chunking a folder of json files containing a HTML test under the key ['hasPart'][0]['text']"
    )

    parser.add_argument(
        "--input_dir",
        type=str,
        required=True,
        help="Path to folder containing the files to chunk",
    )
    parser.add_argument(
        "--max_headers_to_use",
        type=str,
        choices=["h1", "h2", "h3", "h4", "h5"],
        default="h4",
        help="The maximum level of headers to use for chunking",
    )
    parser.add_argument(
        "--hard_max_chunk_word_count",
        type=int,
        default=400,
        help="Hard limit of the size of a chunk (in words).",
    )
    parser.add_argument(
        "--remove_links",
        type=bool,
        default=False,
        help="Whether or not the links should be removed.",
    )
    parser.add_argument(
        "--max_chunk_word_count",
        type=int,
        default=250,
        help="Soft limit of chunk size. Chunks bigger than this limit will be subdivided with lower level headers.",
    )
    parser.add_argument(
        "--min_chunk_word_count",
        type=int,
        default=15,
        help="Minium amount a word a chunk must have. If lower, the chunk is discarded.",
    )

    return parser.parse_args()


def main():
    args = parse_arguments()

    parser = WikitJsonParser()
    chunker = MarkdownChunker(
        max_headers_to_use=args.max_headers_to_use,
        max_chunk_word_count=args.max_chunk_word_count,
        hard_max_chunk_word_count=args.hard_max_chunk_word_count,
        min_chunk_word_count=args.min_chunk_word_count,
    )
    pipeline = WikitJsonPipeline(parser=parser, chunker=chunker)
    pipeline.chunk_directory(args.input_dir)


if __name__ == "__main__":
    main()



class ChunkNorrisException(Exception):
    def __init__(self, message: str):
        pass


class PdfParserException(Exception):
    def __init__(self, message: str):
        pass


class TextNotFoundException(Exception):
    def __init__(self, message: str):
        pass


class PageNotFoundException(Exception):
    def __init__(self, message: str):
        pass



from itertools import groupby
from operator import attrgetter
from copy import deepcopy

import pymupdf  # type: ignore : no stubs

from .components import Cell, TextSpan, PdfTable
from .utils import PdfParserState


class PdfTableExtraction(PdfParserState):
    """Tool that groups methods related to extraction of the tables of a pdf.
    Meant to be as inerited class PdfParser(PdfTableExtraction) as it uses some of
    the attributes of PdfParser, such as self.spans and self.document
    """

    def get_tables(self) -> list[PdfTable]:
        """Parses the table of the document. For this to work, tables
        must have visible "lines".

        Returns:
            PdfTable: the list of tables in the pdf.
        """
        spans_per_page = {
            page: list(spans_on_page)
            for page, spans_on_page in groupby(self.spans, key=lambda span: span.page)
        }
        # get tables on all pages
        tables = []
        for page in self.document.pages(start=self.page_start, stop=self.page_end):  # type: ignore : missing typing in pymupdf -> document.pages() : generator[Page]
            raw_tables: pymupdf.table.TableFinder = page.find_tables(  # type: ignore : missing typing in pymupdf -> Page.find_tables() -> pymupdf.TableFinder
                horizontal_strategy="lines_strict",
                vertical_strategy="lines_strict",
            ).tables
            for raw_table in raw_tables:  # type: ignore : missing typing in pymupdf -> TableFinder.tables  : list[pymupdf.table.Table]
                raw_table = PdfTableExtraction.sanitize_table(raw_table)
                cells = self._get_table_cells(raw_table, spans_per_page[page.number])  # type: ignore : missing typing in pymupdf -> Page.number -> int
                # if the table contains at least 1 span
                if any(cell.spans for cell in cells):
                    tables.append(PdfTable(cells, page.number))  # type: ignore : missing typing in pymupdf -> Page.number -> int

        return sorted(tables, key=attrgetter("order"))

    @staticmethod
    def sanitize_table(
        table: pymupdf.table.Table, wordcount_threshold: int = 60
    ) -> pymupdf.table.Table:
        """Sanitize the table by removing columns that are
        likely to be false detection. Column is considered
        a wrong detection if:
        - all cells are empty
        - one cell as more words than wordcount_threshold

        Args:
            table (pymupdf.table.Table): a pymupdf table
            wordcount_threshold (int) : columns that have at least one cell that contains
                more words than wordcount_threshold will be discarded. Defaults to 60.
        """
        df = table.to_pandas().fillna("")  # type: ignore : missing typing in pymupdf table.to_pandas() -> pandas.DataFrame
        # Get the idx of cols that only have empty cells
        idx_of_cols_to_remove_empty = [
            i for i, col in enumerate(df.columns) if (df[col] != "").sum() == 0  # type: ignore : missing typing in pandas sum()
        ]
        # Get the idx of cols that have at least on cell with a lonf text
        ## Set the header as last raow to include it in word count
        df.loc[-1] = df.columns
        ## Count words in each cell
        wordcount_per_cell = df.map(lambda x: len(str(x).split()))  # type: ignore : missing typing in pandas map()
        idx_of_cols_to_remove_wordcount = [
            i
            for i, col_name in enumerate(wordcount_per_cell.columns)
            if (wordcount_per_cell[col_name] > wordcount_threshold).any()  # type: ignore : missing typing in pandas any()
        ]
        # Get all the x positions of the columns that must be removed
        idx_of_cols_to_remove = set(
            idx_of_cols_to_remove_empty + idx_of_cols_to_remove_wordcount
        )
        ## Get the x0 positions of the columns
        cols_x0_pos = list(set(bbox[0] for bbox in table.cells))
        pos_to_discard = [cols_x0_pos[i] for i in idx_of_cols_to_remove]
        # Remove all cells corresponding to those columns
        table.cells = [
            cell
            for cell in table.cells
            if cell is not None and cell[0] not in pos_to_discard
        ]
        table.header.cells = [  # type: ignore : missing typing in pymupdf header : pymupdf.table.TableHeader
            cell
            for cell in table.header.cells  # type: ignore : missing typing in pymupdf header : pymupdf.table.TableHeader
            if cell is not None and cell[0] not in idx_of_cols_to_remove
        ]
        table.header.names = [  # type: ignore : missing typing in pymupdf header : pymupdf.table.TableHeader
            name
            for i, name in enumerate(table.header.names)  # type: ignore : missing typing in pymupdf header : pymupdf.table.TableHeader
            if i not in idx_of_cols_to_remove
        ]

        return table

    def _get_table_cells(
        self, raw_table: pymupdf.table.Table, spans_on_page: list[TextSpan]
    ) -> list[Cell]:
        """From a pymupdf table,
        builds a list of Cell objects and binds each
        of them the spans that they contain.

        In pymupdf, the "cell" attribute of the Table class
        is just a list of tuple representing (x0, y0, x1, y1)
        coordinates.
        From this we want to build a list of "Cell" objects
        and binds them their corresponding text.

        Args:
            raw_table (pymupdf.table.Table): a table directly extracted from pymupdf's find_table() method.
            spans_on_page (list[TextSpan]): a list of spans that are on the same page of the table.
        """
        cells: list[Cell] = []
        for cell_coords in raw_table.cells:
            cell = Cell(*cell_coords)
            spans_to_bind: list[TextSpan] = []
            for span in spans_on_page:
                small_bbox = PdfTableExtraction._get_smaller_bbox(span.bbox)
                # if the cell contains the span -> bind span to cell
                if cell.contains(small_bbox):  # type: ignore : missing typing in pymupdf -> Rect.contains(r: Point | Rect) -> bool
                    spans_to_bind.append(span)
                # elif cell intersects the span -> span is on multiple cells -> create new splitted span
                elif cell.intersects(small_bbox):  # type: ignore : missing typing in pymupdf -> Rect.intersects(r: Rect) -> bool
                    splitted_span = PdfTableExtraction._split_span(cell, span)
                    spans_to_bind.append(splitted_span)

            cell.spans = spans_to_bind
            cells.append(cell)

        return cells

    @staticmethod
    def _get_smaller_bbox(bbox: pymupdf.Rect, offset: float = 3) -> pymupdf.Rect:
        """Considering a pymupdf.Rect, returns
        a smaller Rect considering the provided offset.
        Example :
            Rect(10, 10, 20 ,20) and offset=2 will return Rect(12, 12, 18, 18)

        If the resulting rectangle is invalid, returns an empty Rectangle centered.
        Example :
            Rect(10, 10, 20, 20) and offset=8 would return Rect(18, 18, 12, 12) which is invalid
            -> Instead it will return Rect(15, 15, 15, 15)

        Args:
            bbox (pymupdf.Rect): the rectangle to consider.
            offset (float): the offset to use.

        Returns:
            pymupdf.Rect : the smaller bbox
        """
        small_bbox = pymupdf.Rect(
            bbox.x0 + offset,  # type: ignore : missing typing in pymuPdf | Rect.x0 : float
            bbox.y0 + offset,  # type: ignore : missing typing in pymuPdf | Rect.y0 : float
            bbox.x1 - offset,  # type: ignore : missing typing in pymuPdf | Rect.x1 : float
            bbox.y1 - offset,  # type: ignore : missing typing in pymuPdf | Rect.y1 : float
        )
        if not small_bbox.is_valid:  # type: ignore : missing typing in pymuPdf | Rect.is_valid : bool
            return pymupdf.Rect(
                (bbox.x0 + bbox.x1) / 2,  # type: ignore : missing typing in pymuPdf : Rect.x0, Rect.y0, Rect.x1, Rect.y1 are all float
                (bbox.y0 + bbox.y1) / 2,  # type: ignore : missing typing in pymuPdf : Rect.x0, Rect.y0, Rect.x1, Rect.y1 are all float
                (bbox.x0 + bbox.x1) / 2,  # type: ignore : missing typing in pymuPdf : Rect.x0, Rect.y0, Rect.x1, Rect.y1 are all float
                (bbox.y0 + bbox.y1) / 2,  # type: ignore : missing typing in pymuPdf : Rect.x0, Rect.y0, Rect.x1, Rect.y1 are all float
            )

        return small_bbox

    @staticmethod
    def _split_span(cell: Cell, span: TextSpan) -> TextSpan:
        """Considering a cell and a span,
        splits the span (or actually creates a new one)
        with only the text contained inside the cell.

        Args:
            cell (Cell): the cell to consider.
            span (TextSpan): the span to split.

        Returns:
            TextSpan: a new span, representing the portion of the provided span
                that is inside the cell.
        """
        # Compute the amount of character that are inside the cell
        intersect = deepcopy(cell).intersect(span.bbox)  # type: ignore : missing typing in pymuPdf | Rect.intersect(r : Rect) -> bool
        x_portion_to_keep = intersect.width / span.bbox.width  # type: ignore : missing typing in pymuPdf | Rect.width : float
        n_char_to_keep = int(round(x_portion_to_keep * len(span.text), 0))
        # Determine which portion of span is in cell (left or right)
        from_left = intersect.x0 == span.bbox.x0  # type: ignore : missing typing in pymuPdf | Rect.x0 : float
        # Get text of span inside cell
        text = (
            span.text[:n_char_to_keep]
            if from_left
            else span.text[len(span.text) - n_char_to_keep :]
        )
        new_span = TextSpan(
            bbox=intersect,
            text=text,
            font=span.font,
            color=span.fontcolor,
            size=span.fontsize,
            flags=span.flags,
            ascender=span.ascender,
            descender=span.descender,
            origin=pymupdf.Point(intersect.x0, span.origin.y),  # type: ignore : missing typing in pymuPdf | Rect.x0 : float, Point.y : float
            page=span.page,
        )
        new_span.order = span.order

        return new_span



from abc import ABC, abstractmethod
from ..chunkers.abstract_chunker import AbstractChunker
from ..chunkers.tools import Chunk
from ..parsers.abstract_parser import AbstractParser


class AbstractPipeline(ABC):
    parser: AbstractParser
    chunker: AbstractChunker

    def __init__(self, parser: AbstractParser, chunker: AbstractChunker) -> None:
        self.parser = parser
        self.chunker = chunker

    @abstractmethod
    def chunk_string(self, string: str) -> list[Chunk]:
        """Given a string, returns a list of chunks.
        Leverages the provided chunker and parser.
        Might use self.parser.parse_string() and self.parser.parse_file()
        along with self.chunker.chunk_string().
        """
        pass

    @abstractmethod
    def chunk_file(self, filepath: str) -> list[Chunk]:
        """Given a filepath, returns a list of chunks.
        Leverages the provided chunker and parser.
        Might use self.parser.parse_string() and self.parser.parse_file()
        along with self.chunker.chunk_string().
        """
        pass

    def __call__(self, string: str) -> list[Chunk]:
        """Gets chunks a string

        Args:
            md_string (str): the markdown string

        Returns:
            Chunks: the list of chunk's texts
        """
        return self.chunk_string(string)

    def save_chunks(
        self, chunks: list[Chunk], output_filename: str, remove_links: bool = False
    ) -> None:
        """Saves the chunks at the designated location
        as a json list of chunks.

        Args:
            chunks (Chunks): the chunks.
            output_filename (str): the JSON file where to save the files. Must be json.
            remove_links (bool): Whether or not links should be remove from the chunk's text content.
        """
        self.chunker.save_chunks(chunks, output_filename, remove_links)



from .components import TextSpan, TextLine, TextBlock, Link, Cell, PdfTable
from .utils import PdfParserState, PdfParserUtilities
from .extract_tables import PdfTableExtraction
from .extract_toc import PdfTocExtraction
from .export import PdfExport
from .plot import PdfPlotter



import os
from collections import Counter
from copy import deepcopy
from itertools import groupby
from pathlib import Path
from typing import Literal
from pydantic import validate_call
import pymupdf  # type: ignore : no stubs

from ...exceptions.exceptions import (
    PageNotFoundException,
    PdfParserException,
    TextNotFoundException,
)
from ...types.types import MarkdownString
from .tools import (
    TextSpan,
    TextLine,
    TextBlock,
    Link,
    PdfPlotter,
    PdfTableExtraction,
    PdfTocExtraction,
    PdfExport,
    PdfParserUtilities,
    PdfParserState,
)


class PdfParser(
    PdfTableExtraction,
    PdfPlotter,
    PdfTocExtraction,
    PdfExport,
    PdfParserUtilities,
    PdfParserState,
):
    """Class that parses the document."""

    extract_tables: bool = True
    add_headers: bool = True
    use_ocr: Literal["always", "auto", "never"] = "auto"
    ocr_language: str = "fra+eng"
    body_line_spacing: float | None = None

    def __init__(
        self,
        *,
        extract_tables: bool = True,
        add_headers: bool = True,
        use_ocr: Literal["always", "auto", "never"] = "auto",
        ocr_language: str = "fra+eng",
        body_line_spacing: float | None = None,
    ) -> None:
        """Initializes a pdf parser
        Args:
            extract_tables (bool, optional): whether or not tables should be extracted.
                Defaults to True.
            add_headers (bool, optional): if True, the parser will try to find a table of content.
                either in documents or in metadata and style the headers accordingly.
                Defaults to True.
            use_ocr (str, optional): whether or not OCR should be used.
                Allows to detect text on images but keep in mind that
                this might include text you actually do not want, such as screenshots.
                Must be one of ["always", "auto", "never"]. Default to "auto".
            ocr_language (str, optional) : the languages to consider for OCR.
                Must be a string of 3 letter codes languages separated by "+".
                Example : "fra+eng+ita"
            body_line_spacing (float, optional) : the size of the space between 2 lines of the body
                of the document. Generally around 1. If None, an automatic method will try to find it.
                Tweak this parameter for better merging of lines into blocks.
        """
        self.add_headers = add_headers
        self.extract_tables = extract_tables
        self.use_ocr = use_ocr
        self.ocr_language = ocr_language
        self.body_line_spacing = body_line_spacing

        if self.use_ocr != "never":
            self.check_ocr_config_is_valid()

    @validate_call
    def parse_file(
        self,
        filepath: str,
        page_start: int = 0,
        page_end: int | None = None,
    ) -> MarkdownString:
        """Parses a pdf document and returns
        its corresponding Markdown formatted string.

        Args:
            filepath (str): the path to the file to parse.
            page_start (int, optional): the page to start parsing from. Defaults to 0.
            page_end (int, optional): the page to stop parsing. None to parse until last page. Defaults to None.

        Returns:
            MarkdownString: The markdown string.
        """
        self.filepath = filepath
        if Path(filepath).suffix != ".pdf":
            raise PdfParserException("Only .pdf files can be passed to PdfParser.")
        self.document = pymupdf.open(filepath, filetype="pdf")
        self._set_page_range(page_start, page_end)

        self.parse_document()

        return self.to_markdown()

    @validate_call
    def parse_string(
        self, string: bytes, page_start: int = 0, page_end: int | None = None
    ) -> MarkdownString:
        """Parses a byte string obtained from a pdf document
        and returns its corresponding Markdown formatted string.

        Args:
            string (bytes): a bytes stream.
            page_start (int, optional): the page to start parsing from. Defaults to 0.
            page_end (int, optional): the page to stop parsing. None to parse until last page. Defaults to None.

        Returns:
            MarkdownString: the markdown formatted string.
        """
        self.document = pymupdf.open(stream=string, filetype="pdf")
        self._set_page_range(page_start, page_end)

        self.parse_document()

        return self.to_markdown()

    def _set_page_range(self, page_start: int, page_end: int | None) -> None:
        """Initializes the self.page_start and self.page_end

        Args:
            page_start (int): the page to start parsing from. Defaults to 0.
            page_end (int | None): the page to stop parsing.
        """
        self.page_start = page_start
        if self.document.page_count == 0:  # type: ignore : missing typing in pymupdf -> document.page_count : int
            raise PageNotFoundException("The provided document contains no pages !")
        if page_end is None:
            self.page_end = self.document.page_count  # type: ignore : missing typing in pymupdf -> document.page_count : int
        else:
            self.page_end = min(page_end, self.document.page_count)  # type: ignore : missing typing in pymupdf -> document.page_count : int
        if self.page_start >= self.page_end:  # type: ignore : missing typing in pymupdf -> document.page_count : int
            raise ValueError("Arg 'page_end' must be greater than 'page_start'.")

    def parse_document(self) -> None:
        """Parses a pdf document.

        Returns:
            str: the pymupdf object
        """
        self.spans = self._create_spans()
        if not self.spans or all(span.is_header_footer for span in self.spans):
            raise TextNotFoundException(
                'No text content found in document. You may want to set use_ocr="always".'
            )
        self.tables = self.get_tables() if self.extract_tables else []
        self.spans = self._flag_table_spans(self.spans)
        self.lines = PdfParser._create_lines(self.spans)
        self.blocks = self._create_blocks(self.lines)
        self.main_title = self._get_document_main_title()
        self.toc = self.get_toc() if self.add_headers else []

    def check_ocr_config_is_valid(self):
        """Check that the OCR configuration is valid"""
        tessdata_location = os.environ.get("TESSDATA_PREFIX")
        if not tessdata_location:
            raise PdfParserException(
                'To use OCR, the "TESSDATA_PREFIX" must be set as environment variable in order to locate traineddata files. For more info see https://pymupdf.readthedocs.io/en/latest/installation.html#enabling-integrated-ocr-support\nYou may otherwise want to deactivate OCR : PdfParser(use_ocr=False).',
            )
        language_list = self.ocr_language.split("+")
        for lang in language_list:
            if not os.path.exists(
                os.path.join(tessdata_location, f"{lang}.traineddata")
            ):
                raise PdfParserException(
                    f"Tesseract's {lang}.traineddata file not found at {tessdata_location}. You might need to download the corresponding file from https://github.com/tesseract-ocr/tessdata and place it in {tessdata_location}",
                )

    def _create_spans(self) -> list[TextSpan]:
        """Prepares the parsed spans

        Returns:
            list[textSpan]: the spans, after preprocessing
        """
        spans = self._extract_spans()
        for i, span in enumerate(spans):
            span.order = i
        spans = self._flag_headers_footers(spans)
        spans = self._bind_links_to_spans(spans)

        return spans

    def _extract_spans(self) -> list[TextSpan]:
        """Get the spans of the pages"""

        spans: list[TextSpan] = []
        for page in self.document.pages(start=self.page_start, stop=self.page_end):  # type: ignore : missing typing in pymupdf -> document.pages() : generator[Page]
            match self.use_ocr:
                case "always":
                    textpage: pymupdf.TextPage = page.get_textpage_ocr(  # type: ignore : missing typing in pymupdf
                        language=self.ocr_language, dpi=72, full=False
                    )
                case "auto":
                    textpage: pymupdf.TextPage = page.get_textpage()  # type: ignore : missing typing in pymupdf
                    page_spans = PdfParser._extract_spans_from_textpage(
                        textpage, page.number  # type: ignore : missing typing in pymupdf -> page.number: int
                    )
                    if not page_spans:
                        textpage: pymupdf.TextPage = page.get_textpage_ocr(  # type: ignore : missing typing in pymupdf
                            language=self.ocr_language, dpi=72, full=False
                        )
                case "never":
                    textpage: pymupdf.TextPage = page.get_textpage()  # type: ignore : missing typing in pymupdf
            spans.extend(PdfParser._extract_spans_from_textpage(textpage, page.number))  # type: ignore : missing typing in pymupdf -> page.number: int

        return spans

    @staticmethod
    def _extract_spans_from_textpage(
        textpage: pymupdf.TextPage, page_number: int
    ) -> list[TextSpan]:
        """Extracts a list of spans from a TextPage

        Args:
            textpage (pymupdf.TextPage): the TextPage to extract the spans from.
            page_number (int): the page number of the provided textpage.

        Returns:
            list[TextSpan]: a list of textspan objects
        """
        spans: list[TextSpan] = []
        page_dict: dict[str, str] = textpage.extractDICT()  # type: ignore : missing typing in pymupdf
        for block in page_dict["blocks"]:
            for line in block["lines"]:
                for span in line["spans"]:
                    spans.append(
                        TextSpan(
                            page=page_number,
                            **span,
                        )
                    )

        return spans

    def _flag_table_spans(self, spans: list[TextSpan]) -> list[TextSpan]:
        """Flags the span if it belongs to any table
        already parsed in the tables.
        Stores the value in span.isin_table attribute

        Args:
            spans (list[TextSpan]) : the list of spans

        Returns:
            list[TextSpan] : the list of spans with added attribute "isin_table"
        """
        for span in spans:
            span.isin_table = any(
                pymupdf.Rect(table.bbox).contains(span.origin)  # type: ignore : missing typing in pymupdf | Rect.contains(x) : bool
                and span.page == table.page
                for table in self.tables
            )

        return spans

    def _flag_headers_footers(self, spans: list[TextSpan]) -> list[TextSpan]:
        """Flags spans that are headers and footers (inplace)
        We consider that if a bbox is present at the same place on a page
        on more that 33% of the pages of the document, it is a header or footer.
        Process:
        - Count the bbox locations of all spans.
        - Remove spans having bbox location count superior to document's page count / 2

        Args:
            spans (TextSpan): the list of spans with attribute is_header_footer updated.

        """
        if not self.document.page_count > 3:  # type: ignore : missing typing in pymupdf | document.page_count : int
            return spans

        bbox_location_counts = Counter((span.bbox) for span in spans)
        header_footer_bboxes = {
            bbox: count
            for bbox, count in bbox_location_counts.items()
            if count > self.document.page_count / 3  # type: ignore : missing typing in pymupdf | document.page_count : int
        }
        for span in spans:
            span.is_header_footer = span.bbox in header_footer_bboxes

        return spans

    def _bind_links_to_spans(self, spans: list[TextSpan]) -> list[TextSpan]:
        """In a pdf, links are just an invisible clickable box
        layered on top of a span.
        This method gets the links of the pdf
        and binds them to their corresponding span.

        Args:
            spans (list[TextSpan]): the list of spans

        Returns:
            list[TextSpan]: the list of spans, with the "link" and "has_link" attributes updated
        """
        spans_per_page_map = {
            page: list(spans_on_page)
            for page, spans_on_page in groupby(spans, key=lambda span: span.page)
        }
        for page_idx, spans_on_page in spans_per_page_map.items():
            page = self.document.load_page(page_idx)  # type: ignore : missing typing in pymupdf | document.load_page(page_id : int) -> Page
            for link in page.links(kinds=(pymupdf.LINK_URI,)):  # type: ignore : missing typing in pymupdf | page.links(kinds : Literal) -> generator
                link = Link(uri=link["uri"], bbox=link["from"])
                corresponding_span_idx = PdfParser._get_span_corresponding_to_link(
                    link, spans_on_page
                )
                if corresponding_span_idx is not None:
                    spans_on_page[corresponding_span_idx].link = link

        return spans

    @staticmethod
    def _get_span_corresponding_to_link(
        link: Link, spans: list[TextSpan]
    ) -> int | None:
        """Given a link and a list of spans, it finds the
        span corresponding to this link. It returns the
        index of that span in the list.

        To find the span a link is bound to, we consider the area of the
        intersection of the link's bbox and the span's bbox.
        Two different situations:
        - 1) there is only one span with a high bbox intersection value with the link
            => this span is the one of the link
        - 2) there are multiple spans with a high bbox intersection.
            => for some reason, the span corresponding to the link is not the one
            with the biggest bbox intersection area, but the one with the second biggest.
        Args:
            link (Link): the link to find the corresponding span to
            spans (list[TextSpan]): a list of spans

        WARNING : MAY return None if no span with bbox intersection with the link's bbox has been found

        Returns:
            int: the index of the corresponding span in the list
        """
        # build list of (span_index, bbox_interect_area)
        intersect_areas: list[tuple[int, float]] = [
            (i, deepcopy(link.bbox).intersect(span.bbox).get_area())  # type: ignore : missing typing in pymupdf | Rect.intersect(r: Rect) -> Rect
            for i, span in enumerate(spans)
        ]
        intersect_areas = sorted(intersect_areas, key=lambda x: x[1], reverse=True)
        best_idx, best_area = intersect_areas[0]
        if len(intersect_areas) == 1:
            return None if best_area == 0 else best_idx
        elif best_area / 2 > intersect_areas[1][1]:
            return best_idx
        else:
            intersect_areas = [item for item in intersect_areas if item[1] < best_area]
            if intersect_areas:
                return intersect_areas[0][0]
            return None

    @staticmethod
    def _create_lines(spans: list[TextSpan]) -> list[TextLine]:
        """Consolidate the consecutive spans by grouping them together
        in a TextLine when possible if they belong to the same line.
        Spans can be merged if :
        - they do not belong to table or header/footer
        - they have the same y positions
        Args:
            spans (list[TextSpan]): the list of spans

        Returns:
            list[TextLine]: the list of lines
        """
        if len(spans) == 0:
            return []

        lines: list[TextLine] = []
        # do not consider footer/header or table spans
        spans_to_merge = [
            span for span in spans if not span.is_header_footer and not span.isin_table
        ]
        # group spans by page
        spans_grouped_per_page = (
            list(spans_on_page)
            for _, spans_on_page in groupby(spans_to_merge, key=lambda span: span.page)
        )
        for spans_on_page in spans_grouped_per_page:
            buffer = [spans_on_page[0]]
            for span in spans_on_page[1:]:
                if span.origin.y == buffer[-1].origin.y or span.is_superscripted:  # type: ignore : missing typing in pymupdf | Point.y : float
                    buffer.append(span)
                else:
                    lines.append(TextLine(buffer))
                    buffer = [span]
            lines.append(TextLine(buffer))

        return lines

    @staticmethod
    def _get_line_spacing(lines: list[TextLine]) -> float:
        """Determine the linespace of the body of the document's content.
        It this case, "linespace" refers to the vertical distance
        between the bboxes of 2 consecutive lines.
        This linespace can be used to merge the lines together into blocks.

        Args:
            lines (list[TextLine]): the parsed lines

        Returns:
            float: the value of the linespace
        """
        linespace_counts = Counter(
            (
                round(curr_line.bbox.y0 - prev_line.bbox.y1, 1)  # type: ignore : missing typing in pymupdf | Rect.y0 : float
                for curr_line, prev_line in zip(lines[1:], lines[:-1])
            )
        )

        return max(linespace_counts, key=linespace_counts.get) + 0.2

    def _create_blocks(self, lines: list[TextLine]) -> list[TextBlock]:
        """Groups lines together into blocks.
        A block is a group of lines the is not separated by extra spacing.
        For example, it can be a paragraph, or the title of a section.

        Args:
            lines (list[TextLine]): the lines to group
            body_linespacing (float): the distance between the bboxes of two
                consecutive lines below which 2 lines are considered belonging to the same block

        Returns:
            list[TextBlock]: the blocks
        """
        if len(lines) == 0:
            return []
        if self.body_line_spacing is None:
            self.body_line_spacing = PdfParser._get_line_spacing(lines)

        blocks: list[TextBlock] = []
        buffer = [lines[0]]
        for line in lines[1:]:
            # if previous line was emtpy ("\n") => new block
            # or if new line "far away" from previous line => new block
            if buffer[-1].is_empty or (
                line.bbox.y0 - self.body_line_spacing > buffer[-1].bbox.y1  # type: ignore : missing typing in pymupdf | Rect.y0 : float
            ):
                blocks.append(TextBlock(buffer))
                buffer = [line]
            # new line is close to previous line
            else:
                # if new line is far up on page than previous line => change page or multicolumn => new block
                if line.bbox.y1 <= buffer[-1].bbox.y0:  # type: ignore : missing typing in pymupdf | Rect.y0 : float
                    blocks.append(TextBlock(buffer))
                    buffer = [line]
                # Two consecutive lines belong to the same block
                else:
                    buffer.append(line)
        blocks.append(TextBlock(buffer))

        return blocks

    def cleanup_memory(self):
        """Cleans up memory"""
        self.document.close()
        self._document = None
        self.filepath = None
        self.page_start = 0
        self.page_end = None
        self.spans = []
        self.lines = []
        self.blocks = []
        self.tables = []
        self.main_title = ""



from typing import Literal
import pymupdf  # type: ignore : no stubs

from ....exceptions.exceptions import PdfParserException
from .components import PdfTable, TextBlock, TextLine, TextSpan


class PdfParserState:
    """Contains the state of the class.
    Intended to be inherited by all pdfParser components
    """

    _document: pymupdf.Document | None = None
    filepath: str | None = None
    page_start: int = 0
    page_end: int | None = None
    spans: list[TextSpan] = []
    lines: list[TextLine] = []
    blocks: list[TextBlock] = []
    tables: list[PdfTable] = []
    main_title: str = ""

    @property
    def document(self) -> pymupdf.Document:
        if self._document is None:
            raise PdfParserException(
                "No document parsed. You might call parser.parse_file('my_file.pdf')."
            )
        return self._document

    @document.setter
    def document(self, doc: pymupdf.Document) -> None:
        self._document = doc


class PdfParserUtilities(PdfParserState):

    def get_document_orientation(self) -> Literal["portrait", "landscape"]:
        """Determines the orientation of the document
        using the first page of the document.

        Returns:
            str: the orientation "landscape" or "portrait"
        """
        page_dims: pymupdf.Rect = self.document[0].rect

        return "portrait" if page_dims.height > page_dims.width else "landscape"  # type: ignore missing typing in pymupdf | Rect.height : float, Rect.width : float



from abc import ABC, abstractmethod
import json
import os
from typing import Any

from pydantic import validate_call

from .tools.tools import Chunk


class AbstractChunker(ABC):

    @validate_call
    @abstractmethod
    def chunk_string(self, string: Any) -> list[Chunk]:
        """Considering the output of a specified parser,
        must contain all the logic for parsing a string.

        One might use the output of the parser: string.content

        Args:
            string (Any): the string output from the
                provided parser. Can be any type that inherits
                AbstractInOutType.

        Returns:
            list[Chunk]: A list of chunks
        """
        pass

    def __call__(self, string: Any) -> list[Chunk]:
        """Shortcut to call chunk_string() method."""
        return self.chunk_string(string)

    def save_chunks(
        self, chunks: list[Chunk], output_filename: str, remove_links: bool = False
    ) -> None:
        """Saves the chunks in .md files at the designated location.

        Args:
            chunks (Chunks): the chunks.
            output_filename (str): the JSON file where to save the files. Must be json.
            remove_links (bool): Whether or not links should be remove from the chunk's text content.
        """
        directory = os.path.dirname(os.path.abspath(output_filename))
        if not os.path.exists(directory):
            os.makedirs(directory)
        content = [
            {
                "text": chunk.get_text(remove_links=remove_links),
                "start_page": chunk.start_page,
                "end_page": chunk.end_page,
                "start_line": chunk.start_line,
            }
            for chunk in chunks
        ]
        with open(output_filename, "w", encoding="utf8") as f:
            json.dump(content, f, ensure_ascii=False)



from argparse import ArgumentParser
import logging
import os

from .chunkers import MarkdownChunker
from .parsers import MarkdownParser, WikitJsonParser, HTMLParser, PdfParser
from .pipelines import BasePipeline, PdfPipeline, WikitJsonPipeline

LOGGER = logging.getLogger(name="ChunkNorris")
logging.basicConfig(level=logging.INFO)


def parse_arguments():
    """Parse the given command-line arguments."""

    parser = ArgumentParser(description="Tool meant to chunk the content of a file")
    parser.add_argument(
        "--filepath",
        type=str,
        required=True,
        help="Path to the file to chunk",
    )
    parser.add_argument(
        "--max_headers_to_use",
        type=str,
        choices=["h1", "h2", "h3", "h4", "h5"],
        default="h4",
        help="The maximum level of titles to use for chunking",
    )
    parser.add_argument(
        "--max_chunk_word_count",
        type=int,
        default=250,
        help="Soft limit of chunk size (in words). Chunks bigger than this limit will be subdivided with lower level headers if available.",
    )
    parser.add_argument(
        "--hard_max_chunk_word_count",
        type=int,
        default=400,
        help="Hard limit of chunk size (in words). Chunks bigger than that will be subdivided so that each subchunk has less words than specified value.",
    )
    parser.add_argument(
        "--min_chunk_word_count",
        type=int,
        default=15,
        help="Minium amount a word a chunk must have. If lower, the chunk is discarded.",
    )
    parser.add_argument(
        "--remove_links",
        type=bool,
        default=False,
        help="Whether or not the links should be removed.",
    )
    parser.add_argument(
        "--use_ocr",
        type=str,
        choices=["auto", "never", "always"],
        default="auto",
        help="For PDF only : whether or not OCR should be used.",
    )
    parser.add_argument(
        "--ocr_language",
        type=str,
        default="fra+eng",
        help='For PDF only : the languages to consider for OCR. Must be a string of 3 letter codes languages separated by "+", such as "eng+fra"',
    )
    parser.add_argument(
        "--extract_tables",
        type=bool,
        default=True,
        help="For PDF only : whether or not tables should be extracted",
    )
    return parser.parse_args()


def main():
    args = parse_arguments()

    filename, fileext = os.path.splitext(args.filepath)

    chunker = MarkdownChunker(
        max_headers_to_use=args.max_headers_to_use,
        max_chunk_word_count=args.max_chunk_word_count,
        hard_max_chunk_word_count=args.hard_max_chunk_word_count,
        min_chunk_word_count=args.min_chunk_word_count,
    )

    LOGGER.info("Chunking...")

    match fileext:
        case ".md":
            parser = MarkdownParser()
            pipeline = BasePipeline(parser, chunker)
        case ".html":
            parser = HTMLParser()
            pipeline = BasePipeline(parser, chunker)
        case ".pdf":
            parser = PdfParser(
                extract_tables=args.extract_tables,
                ocr_language=args.ocr_language,
                use_ocr=args.use_ocr,
            )
            pipeline = PdfPipeline(parser, chunker)
        case ".json":
            parser = WikitJsonParser()
            pipeline = WikitJsonPipeline(parser, chunker)
        case _:
            raise ValueError(
                "ChunkNorris currently only supports .md, .html, .pdf and .json files."
            )

    chunks = pipeline.chunk_file(filepath=args.filepath)
    LOGGER.info("%i obtained !", len(chunks))

    args.output_filename = filename + "-chunks.json"
    pipeline.save_chunks(chunks, args.output_filename, args.remove_links)
    LOGGER.info("Chunks saved at %s", args.output_filename)


if __name__ == "__main__":
    main()
